"""
Google OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/google.html
"""

from .google import GoogleOAuth2
from .open_id_connect import OpenIdConnectAuth


class GoogleOpenIdConnect(GoogleOAuth2, OpenIdConnectAuth):
    name = "google-openidconnect"
    OIDC_ENDPOINT = "https://accounts.google.com"
    # differs from value in discovery document
    # http://openid.net/specs/openid-connect-core-1_0.html#rfc.section.15.6.2
    ID_TOKEN_ISSUER = "accounts.google.com"

    def user_data(self, access_token, *args, **kwargs):
        """Return user data from Google API"""
        return self.get_json(
            "https://openidconnect.googleapis.com/v1/userinfo",
            params={"access_token": access_token, "alt": "json"},
        )



import json

import responses

from social_core.tests.backends.oauth import BaseAuthUrlTestMixin, OAuth2Test

from .open_id_connect import OpenIdConnectTest

JWK_KEY = {
    "kty": "RSA",
    "d": "ZmswNokEvBcxW_Kvcy8mWUQOQCBdGbnM0xR7nhvGHC-Q24z3XAQWlMWbsmGc_R1o"
    "_F3zK7DBlc3BokdRaO1KJirNmnHCw5TlnBlJrXiWpFBtVglUg98-4sRRO0VWnGXK"
    "JPOkBQ6b_DYRO3b0o8CSpWowpiV6HB71cjXTqKPZf-aXU9WjCCAtxVjfIxgQFu5I"
    "-G1Qah8mZeY8HK_y99L4f0siZcbUoaIcfeWBhxi14ODyuSAHt0sNEkhiIVBZE7QZ"
    "m-SEP1ryT9VAaljbwHHPmg7NC26vtLZhvaBGbTTJnEH0ZubbN2PMzsfeNyoCIHy4"
    "4QDSpQDCHfgcGOlHY_t5gQ",
    "e": "AQAB",
    "use": "sig",
    "kid": "testkey",
    "alg": "RS256",
    "n": "pUfcJ8WFrVue98Ygzb6KEQXHBzi8HavCu8VENB2As943--bHPcQ-nScXnrRFAUg8"
    "H5ZltuOcHWvsGw_AQifSLmOCSWJAPkdNb0w0QzY7Re8NrPjCsP58Tytp5LicF0Ao"
    "Ag28UK3JioY9hXHGvdZsWR1Rp3I-Z3nRBP6HyO18pEgcZ91c9aAzsqu80An9X4DA"
    "b1lExtZorvcd5yTBzZgr-MUeytVRni2lDNEpa6OFuopHXmg27Hn3oWAaQlbymd4g"
    "ifc01oahcwl3ze2tMK6gJxa_TdCf1y99Yq6oilmVvZJ8kwWWnbPE-oDmOVPVnEyT"
    "vYVCvN4rBT1DQ-x0F1mo2Q",
}

JWK_PUBLIC_KEY = {key: value for key, value in JWK_KEY.items() if key != "d"}


class OktaOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.okta.OktaOAuth2"
    user_data_url = "https://dev-000000.oktapreview.com/oauth2/v1/userinfo"
    expected_username = "foo"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "family_name": "Bar",
            "sub": "101010101010101010101",
            "locale": "en",
            "email_verified": True,
            "given_name": "Foo",
            "email": "foo@bar.com",
            "name": "Foo Bar",
            "nickname": "foobar",
            "middle_name": "",
            "profile": "https://example.com/foo.bar",
            "zoneinfo": "America/Los_Angeles",
            "updated_at": 1311280970,
            "preferred_username": "foo",
        }
    )

    def test_login(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_OKTA_OAUTH2_API_URL": "https://dev-000000.oktapreview.com/oauth2"
            }
        )
        self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_OKTA_OAUTH2_API_URL": "https://dev-000000.oktapreview.com/oauth2"
            }
        )
        self.do_partial_pipeline()


class OktaOpenIdConnectTest(OpenIdConnectTest):
    backend_path = "social_core.backends.okta_openidconnect.OktaOpenIdConnect"
    user_data_url = "https://dev-000000.oktapreview.com/oauth2/v1/userinfo"
    issuer = "https://dev-000000.oktapreview.com/oauth2"
    openid_config_body = json.dumps(
        {
            "issuer": "https://dev-000000.oktapreview.com/oauth2",
            "authorization_endpoint": "https://dev-000000.oktapreview.com/oauth2/v1/authorize",
            "token_endpoint": "https://dev-000000.oktapreview.com/oauth2/v1/token",
            "userinfo_endpoint": "https://dev-000000.oktapreview.com/oauth2/v1/userinfo",
            "jwks_uri": "https://dev-000000.oktapreview.com/oauth2/v1/keys",
            "response_types_supported": [
                "code",
                "token",
                "id_token",
                "code token",
                "code id_token",
                "token id_token",
                "code token id_token",
                "none",
            ],
            "subject_types_supported": [
                "public",
            ],
            "id_token_signing_alg_values_supported": [
                "RS256",
            ],
            "scopes_supported": [
                "openid",
                "email",
                "profile",
            ],
            "token_endpoint_auth_methods_supported": [
                "client_secret_post",
                "client_secret_basic",
            ],
            "claims_supported": [
                "aud",
                "email",
                "email_verified",
                "exp",
                "family_name",
                "given_name",
                "iat",
                "iss",
                "locale",
                "name",
                "picture",
                "sub",
            ],
        }
    )
    user_data_body = json.dumps(
        {
            "family_name": "Bar",
            "sub": "101010101010101010101",
            "locale": "en",
            "email_verified": True,
            "given_name": "Foo",
            "email": "foo@bar.com",
            "name": "Foo Bar",
            "nickname": "foobar",
            "middle_name": "",
            "profile": "https://example.com/foo.bar",
            "zoneinfo": "America/Los_Angeles",
            "updated_at": 1311280970,
            "preferred_username": "foo",
        }
    )
    expected_username = "foo"

    def extra_settings(self):
        settings = super().extra_settings()
        # Settings for Okta
        oidc_config = json.loads(self.openid_config_body)
        settings.update(
            {
                "SOCIAL_AUTH_OKTA_OPENIDCONNECT_API_URL": "https://dev-000000.oktapreview.com/oauth2",
                "SOCIAL_AUTH_OKTA_OPENIDCONNECT_OIDC_ENDPOINT": "https://dev-000000.oktapreview.com/oauth2",
                "SOCIAL_AUTH_OKTA_OPENIDCONNECT_JWKS_URI": oidc_config.get("jwks_uri"),
                "SOCIAL_AUTH_OKTA_OPENIDCONNECT_ID_TOKEN_ISSUER": oidc_config.get(
                    "issuer"
                ),
            }
        )
        return settings

    def setUp(self):
        responses.add(
            responses.GET,
            # Note: okta.py strips the /oauth2 prefix using urljoin with absolute path
            "https://dev-000000.oktapreview.com/.well-known/openid-configuration?client_id=a-key",
            status=200,
            body=self.openid_config_body,
            content_type="application/json",
        )
        oidc_config = json.loads(self.openid_config_body)

        responses.add(
            responses.GET,
            oidc_config.get("jwks_uri"),
            status=200,
            json={"keys": [JWK_PUBLIC_KEY]},
        )

        super().setUp()

    def pre_complete_callback(self, start_url):
        super().pre_complete_callback(start_url)
        responses.add(
            responses.GET,
            url=self.backend.userinfo_url(),
            status=200,
            json={"preferred_username": self.expected_username},
        )

    def test_everything_works(self):
        self.do_login()



from .oauth import BaseOAuth2


class DigitalOceanOAuth(BaseOAuth2):
    """
    DigitalOcean OAuth authentication backend.

    Docs: https://developers.digitalocean.com/documentation/oauth/
    """

    name = "digitalocean"
    AUTHORIZATION_URL = "https://cloud.digitalocean.com/v1/oauth/authorize"
    ACCESS_TOKEN_URL = "https://cloud.digitalocean.com/v1/oauth/token"
    SCOPE_SEPARATOR = " "
    EXTRA_DATA = [("expires_in", "expires_in")]

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return response["account"].get("uuid")

    def get_user_details(self, response):
        """Return user details from DigitalOcean account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("name") or ""
        )

        return {
            "username": response["account"].get("email"),
            "email": response["account"].get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, token, *args, **kwargs):
        """Loads user data from service"""
        url = "https://api.digitalocean.com/v2/account"
        auth_header = {"Authorization": f"Bearer {token}"}
        try:
            return self.get_json(url, headers=auth_header)
        except ValueError:
            return None



"""
Github Enterprise OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html
"""

from urllib.parse import urljoin

from ..utils import append_slash
from .github import GithubOAuth2, GithubOrganizationOAuth2, GithubTeamOAuth2


class GithubEnterpriseMixin:
    def api_url(self):
        return append_slash(self.setting("API_URL"))  # type: ignore[reportAttributeAccessIssue]

    def authorization_url(self):
        return self._url("login/oauth/authorize")

    def access_token_url(self):
        return self._url("login/oauth/access_token")

    def _url(self, path):
        return urljoin(append_slash(self.setting("URL")), path)  # type: ignore[reportAttributeAccessIssue]


class GithubEnterpriseOAuth2(GithubEnterpriseMixin, GithubOAuth2):
    """Github Enterprise OAuth authentication backend"""

    name = "github-enterprise"


class GithubEnterpriseOrganizationOAuth2(
    GithubEnterpriseMixin, GithubOrganizationOAuth2
):
    """Github Enterprise OAuth2 authentication backend for
    organizations"""

    name = "github-enterprise-org"
    DEFAULT_SCOPE = ["read:org"]


class GithubEnterpriseTeamOAuth2(GithubEnterpriseMixin, GithubTeamOAuth2):
    """Github Enterprise OAuth2 authentication backend for teams"""

    name = "github-enterprise-team"
    DEFAULT_SCOPE = ["read:org"]



import json
from typing import cast

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest

OIDC_CONFIG_ACADEMIC_ID = """
    {
      "issuer": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud",
      "authorization_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/auth",
      "token_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/token",
      "introspection_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/token/introspect",
      "userinfo_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/userinfo",
      "end_session_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/logout",
      "frontchannel_logout_session_supported": true,
      "frontchannel_logout_supported": true,
      "jwks_uri": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/certs",
      "check_session_iframe": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/login-status-iframe.html",
      "grant_types_supported": [
        "authorization_code",
        "implicit",
        "refresh_token",
        "password",
        "client_credentials",
        "urn:openid:params:grant-type:ciba",
        "urn:ietf:params:oauth:grant-type:device_code"
      ],
      "acr_values_supported": [
        "0",
        "1"
      ],
      "response_types_supported": [
        "code",
        "none",
        "id_token",
        "token",
        "id_token token",
        "code id_token",
        "code token",
        "code id_token token"
      ],
      "subject_types_supported": [
        "public",
        "pairwise"
      ],
      "id_token_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "id_token_encryption_alg_values_supported": [
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5"
      ],
      "id_token_encryption_enc_values_supported": [
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512"
      ],
      "userinfo_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512",
        "none"
      ],
      "userinfo_encryption_alg_values_supported": [
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5"
      ],
      "userinfo_encryption_enc_values_supported": [
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512"
      ],
      "request_object_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512",
        "none"
      ],
      "request_object_encryption_alg_values_supported": [
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5"
      ],
      "request_object_encryption_enc_values_supported": [
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512"
      ],
      "response_modes_supported": [
        "query",
        "fragment",
        "form_post",
        "query.jwt",
        "fragment.jwt",
        "form_post.jwt",
        "jwt"
      ],
      "registration_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/clients-registrations/openid-connect",
      "token_endpoint_auth_methods_supported": [
        "private_key_jwt",
        "client_secret_basic",
        "client_secret_post",
        "tls_client_auth",
        "client_secret_jwt"
      ],
      "token_endpoint_auth_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "introspection_endpoint_auth_methods_supported": [
        "private_key_jwt",
        "client_secret_basic",
        "client_secret_post",
        "tls_client_auth",
        "client_secret_jwt"
      ],
      "introspection_endpoint_auth_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "authorization_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "authorization_encryption_alg_values_supported": [
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA1_5"
      ],
      "authorization_encryption_enc_values_supported": [
        "A256GCM",
        "A192GCM",
        "A128GCM",
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512"
      ],
      "claims_supported": [
        "aud",
        "sub",
        "iss",
        "auth_time",
        "name",
        "given_name",
        "family_name",
        "preferred_username",
        "email",
        "acr"
      ],
      "claim_types_supported": [
        "normal"
      ],
      "claims_parameter_supported": true,
      "scopes_supported": [
        "initials",
        "address",
        "goeidToSub",
        "MPGR_entryUUID_as_preferred_username",
        "org",
        "userservices",
        "accounttype",
        "oxUserId",
        "basic",
        "employeenumber",
        "syncAndShareInstance",
        "phone",
        "roles",
        "gwdg-internal-role",
        "microprofile-jwt",
        "oxContextId",
        "emailCS",
        "goesternMatrikelnummer",
        "vpngroupsNetworkroles",
        "subVPNUserType",
        "acr",
        "owncloud",
        "displayName_to_preferred_username",
        "usRoles",
        "profile",
        "goeId",
        "offline_access",
        "userContextId",
        "goesternSAMAccountName",
        "goesternQuellSystem",
        "userServices_to_authorities",
        "memberOfGoeId",
        "openid",
        "email",
        "nfdi_attributes",
        "memberofdisplayname",
        "syncAttribute1",
        "prefuid",
        "AuthnContextClassRef",
        "web-origins",
        "mailAddresses",
        "uniElectionHash"
      ],
      "request_parameter_supported": true,
      "request_uri_parameter_supported": true,
      "require_request_uri_registration": true,
      "code_challenge_methods_supported": [
        "plain",
        "S256"
      ],
      "tls_client_certificate_bound_access_tokens": true,
      "revocation_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/revoke",
      "revocation_endpoint_auth_methods_supported": [
        "private_key_jwt",
        "client_secret_basic",
        "client_secret_post",
        "tls_client_auth",
        "client_secret_jwt"
      ],
      "revocation_endpoint_auth_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "HS256",
        "HS512",
        "ES256",
        "RS256",
        "HS384",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "backchannel_logout_supported": true,
      "backchannel_logout_session_supported": true,
      "device_authorization_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/auth/device",
      "backchannel_token_delivery_modes_supported": [
        "poll",
        "ping"
      ],
      "backchannel_authentication_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/ext/ciba/auth",
      "backchannel_authentication_request_signing_alg_values_supported": [
        "PS384",
        "RS384",
        "EdDSA",
        "ES384",
        "ES256",
        "RS256",
        "ES512",
        "PS256",
        "PS512",
        "RS512"
      ],
      "require_pushed_authorization_requests": false,
      "pushed_authorization_request_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/ext/par/request",
      "mtls_endpoint_aliases": {
        "token_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/token",
        "revocation_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/revoke",
        "introspection_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/token/introspect",
        "device_authorization_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/auth/device",
        "registration_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/clients-registrations/openid-connect",
        "userinfo_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/userinfo",
        "pushed_authorization_request_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/ext/par/request",
        "backchannel_authentication_endpoint": "https://keycloak.sso.gwdg.de/auth/realms/academiccloud/protocol/openid-connect/ext/ciba/auth"
      },
      "authorization_response_iss_parameter_supported": true
    }
"""

OIDC_CONFIG_DIDMOS = """
    {
      "version": "3.0",
      "token_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post"
      ],
      "claims_parameter_supported": true,
      "request_parameter_supported": false,
      "request_uri_parameter_supported": false,
      "require_request_uri_registration": false,
      "grant_types_supported": [
        "authorization_code",
        "implicit"
      ],
      "frontchannel_logout_supported": false,
      "frontchannel_logout_session_supported": false,
      "backchannel_logout_supported": false,
      "backchannel_logout_session_supported": false,
      "issuer": "https://auth.didmos.nfdi-aai.de",
      "authorization_endpoint": "https://auth.didmos.nfdi-aai.de/OIDC/authorization",
      "jwks_uri": "https://auth.didmos.nfdi-aai.de/OIDC/jwks",
      "response_types_supported": [
        "code",
        "id_token token"
      ],
      "id_token_signing_alg_values_supported": [
        "RS256"
      ],
      "response_modes_supported": [
        "fragment",
        "query"
      ],
      "subject_types_supported": [
        "public"
      ],
      "claim_types_supported": [
        "normal"
      ],
      "claims_supported": [
        "email",
        "sub",
        "preferred_username",
        "name",
        "given_name",
        "family_name",
        "schac_home_organization",
        "eduperson_scoped_affiliation",
        "voperson_external_affiliation",
        "eduperson_assurance",
        "voperson_id",
        "eduperson_unique_id",
        "voperson_verified_email",
        "entitlements",
        "voperson_policy_agreement",
        "orcid",
        "voperson_external_id",
        "eduperson_principal_name",
        "eduperson_target_id",
        "session_id"
      ],
      "scopes_supported": [
        "eduperson_assurance",
        "schac_home_organization",
        "profile",
        "voperson_policy_agreement",
        "voperson_external_affiliation",
        "eduperson_target_id",
        "eduperson_unique_id",
        "voperson_id",
        "email",
        "voperson_verified_email",
        "entitlements",
        "openid",
        "voperson_external_id",
        "eduperson_scoped_affiliation",
        "orcid"
      ],
      "token_endpoint": "https://auth.didmos.nfdi-aai.de/OIDC/token",
      "registration_endpoint": "https://auth.didmos.nfdi-aai.de/OIDC/registration",
      "end_session_endpoint": "https://auth.didmos.nfdi-aai.de/didmos/logout",
      "code_challenge_methods_supported": [
        "S256",
        "S384",
        "S512"
      ],
      "userinfo_endpoint": "https://auth.didmos.nfdi-aai.de/OIDC/userinfo"
    }
"""

OIDC_CONFIG_REGAPP = """
    {
      "response_types_supported": [
        "code",
        "id_token",
        "code id_token"
      ],
      "request_uri_parameter_supported": true,
      "introspection_endpoint": "https://regapp.nfdi-aai.de/oidc/realms/nfdi/protocol/openid-connect/tokeninfo",
      "grant_types_supported": [
        "authorization_code",
        "refresh_token"
      ],
      "scopes_supported": [
        "openid",
        "profile",
        "email"
      ],
      "issuer": "https://regapp.nfdi-aai.de/oidc/realms/nfdi",
      "authorization_endpoint": "https://regapp.nfdi-aai.de/oidc/realms/nfdi/protocol/openid-connect/auth",
      "userinfo_endpoint": "https://regapp.nfdi-aai.de/oidc/realms/nfdi/protocol/openid-connect/userinfo",
      "claims_supported": [
        "sub",
        "iss",
        "aud",
        "mail",
        "name"
      ],
      "jwks_uri": "https://regapp.nfdi-aai.de/oidc/realms/nfdi/protocol/openid-connect/certs",
      "subject_types_supported": [
        "pairwise",
        "public"
      ],
      "id_token_signing_alg_values_supported": [
        "RS256"
      ],
      "token_endpoint_auth_methods_supported": [
        "client_secret_post",
        "private_key_jwt",
        "client_secret_basic"
      ],
      "response_modes_supported": [
        "query",
        "fragment"
      ],
      "token_endpoint": "https://regapp.nfdi-aai.de/oidc/realms/nfdi/protocol/openid-connect/token"
    }
"""

OIDC_CONFIG_UNITY = """
    {
      "authorization_endpoint": "https://login.helmholtz.de/oauth2-as/oauth2-authz",
      "token_endpoint": "https://login.helmholtz.de/oauth2/token",
      "introspection_endpoint": "https://login.helmholtz.de/oauth2/introspect",
      "revocation_endpoint": "https://login.helmholtz.de/oauth2/revoke",
      "issuer": "https://login.helmholtz.de/oauth2",
      "jwks_uri": "https://login.helmholtz.de/oauth2/jwk",
      "scopes_supported": [
        "openid",
        "display_name",
        "sn",
        "single-logout",
        "offline_access",
        "voperson_id",
        "voperson_external_affiliation",
        "entitlements",
        "org_domain",
        "email",
        "profile",
        "credentials",
        "eduperson_scoped_affiliation",
        "eduperson_entitlement",
        "eduperson_principal_name",
        "eduperson_unique_id",
        "eduperson_assurance",
        "sys:scim:read_profile",
        "sys:scim:read_memberships",
        "sys:scim:read_self_group"
      ],
      "response_types_supported": [
        "code",
        "token",
        "id_token",
        "code id_token",
        "id_token token",
        "code token",
        "code id_token token"
      ],
      "response_modes_supported": [
        "query",
        "fragment"
      ],
      "grant_types_supported": [
        "authorization_code",
        "implicit"
      ],
      "code_challenge_methods_supported": [
        "plain",
        "S256"
      ],
      "request_uri_parameter_supported": true,
      "subject_types_supported": [
        "public"
      ],
      "userinfo_endpoint": "https://login.helmholtz.de/oauth2/userinfo",
      "id_token_signing_alg_values_supported": [
        "RS256",
        "ES256"
      ]
    }
"""

OIDC_CONFIG_UNITY_PUNCH = """
    {
      "authorization_endpoint": "https://login.helmholtz.de/punch-oauth2-as/oauth2-authz",
      "token_endpoint": "https://login.helmholtz.de/punch-oauth2/token",
      "introspection_endpoint": "https://login.helmholtz.de/punch-oauth2/introspect",
      "revocation_endpoint": "https://login.helmholtz.de/punch-oauth2/revoke",
      "issuer": "https://login.helmholtz.de/punch-oauth2",
      "jwks_uri": "https://login.helmholtz.de/punch-oauth2/jwk",
      "scopes_supported": [
        "openid",
        "display_name",
        "sn",
        "single-logout",
        "offline_access",
        "email",
        "profile",
        "credentials",
        "eduperson_scoped_affiliation",
        "eduperson_entitlement",
        "eduperson_principal_name",
        "eduperson_unique_id",
        "eduperson_assurance",
        "sys:scim:read_profile",
        "sys:scim:read_memberships",
        "sys:scim:read_self_group"
      ],
      "response_types_supported": [
        "code",
        "token",
        "id_token",
        "code id_token",
        "id_token token",
        "code token",
        "code id_token token"
      ],
      "response_modes_supported": [
        "query",
        "fragment"
      ],
      "grant_types_supported": [
        "authorization_code",
        "implicit"
      ],
      "code_challenge_methods_supported": [
        "plain",
        "S256"
      ],
      "request_uri_parameter_supported": true,
      "subject_types_supported": [
        "public"
      ],
      "userinfo_endpoint": "https://login.helmholtz.de/punch-oauth2/userinfo",
      "id_token_signing_alg_values_supported": [
        "RS256",
        "ES256"
      ]
    }
"""


class NFDIOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.nfdi.HelmholtzOpenIdConnect"
    issuer = "https://login.helmholtz.de/oauth2"
    openid_config_body = OIDC_CONFIG_UNITY
    expected_username = "donald"
    user_data_url = "https://login.helmholtz.de/oauth2/userinfo"
    user_data_body = json.dumps(
        {
            "display_name": "Donald Duck",
            "eduperson_assurance": [
                "https://refeds.org/assurance",
                "https://refeds.org/assurance/ID/unique",
                "https://refeds.org/assurance/ID/eppn-unique-no-reassign",
                "https://refeds.org/assurance/ATP/ePA-1d",
                "https://refeds.org/assurance/ATP/ePA-1m",
                "https://refeds.org/assurance/IAP/local-enterprise",
                "https://refeds.org/assurance/IAP/low",
                "https://refeds.org/assurance/IAP/medium",
                "https://refeds.org/assurance/profile/cappuccino",
                "https://aarc-project.eu/policy/authn-assurance/assam",
            ],
            "eduperson_entitlement": [
                "urn:mace:dir:entitlement:common-lib-terms",
                "http://bwidm.de/entitlement/bwLSDF-SyncShare",
                "urn:canard:mouseton.edu:group:KIT#login.mouseton.edu",
                "urn:canard:h-df.de:group:m-team:feudal-developers#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:Arbeitskreise#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:HIFIS:Associates#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:Arbeitskreise:AG IT Services#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:Helmholtz-member#login.mouseton.edu",
                "urn:canard:h-df.de:group:m-team#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:Helmholtz-all#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:HIFIS#login.mouseton.edu",
            ],
            "eduperson_principal_name": "lo0018@duckburg.edu",
            "eduperson_scoped_affiliation": [
                "employee@login.mouseton.edu",
                "member@login.mouseton.edu",
            ],
            "eduperson_unique_id": "42234223422342234223422342234223@login.mouseton.edu",
            "email": "donald.hardt@duckburg.edu",
            "email_verified": True,
            "entitlements": [
                "urn:mace:dir:entitlement:common-lib-terms",
                "http://duckburg.edu/entitlement/sync-and-share",
                "urn:canard:mouseton.edu:group:KIT#login.mouseton.edu",
                "urn:canard:duckburg.edu:group:m-team:developers#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:Workgroup:IT Services#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:mouseton-member#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:m-team#login.mouseton.edu",
                "urn:canard:mouseton.edu:group:mouseton-all#login.mouseton.edu",
            ],
            "family_name": "Duck",
            "given_name": "Donald",
            "iss": "https://login.mouseton.edu/oauth2",
            "name": "Donald Duck",
            "org_domain": "duckburg.edu",
            "preferred_username": "donald",
            "sn": "Duck",
            "ssh_public_key": "ssh-ed25519 AAAAC3N4224224224224224224224224224224224224224224224224224224223ym/ donald@home\n",
            "sub": "42234223-4223-4223-4223-422342234223",
            "voperson_external_affiliation": [
                "employee@duckburg.edu",
                "faculty@duckburg.edu",
                "member@duckburg.edu",
            ],
            "voperson_id": "42234223422342234223422342234223@login.mouseton.edu",
        }
    )

    def test_do_not_override_endpoint(self):
        self.backend.OIDC_ENDPOINT = self.issuer
        self.assertEqual(self.backend.oidc_endpoint(), self.issuer)

    def test_entitlements_empty(self):
        self.assertEqual(self.backend.entitlement_allowed([]), True)

    def test_entitlements_allowed(self):
        self.backend.ALLOWED_ENTITLEMENTS = ["foo", "baz"]
        self.assertEqual(self.backend.entitlement_allowed(["foo", "bar"]), True)

    def test_entitlements_not_allowed(self):
        self.backend.ALLOWED_ENTITLEMENTS = ["baz"]
        self.assertEqual(self.backend.entitlement_allowed(["foo"]), False)

    def test_get_user_details(self):
        testdata = self.backend.get_user_details(
            json.loads(cast("str", self.user_data_body))
        )
        self.assertEqual(testdata["username"], "donald")
        self.assertEqual(testdata["email"], "donald.hardt@duckburg.edu")
        self.assertEqual(testdata["fullname"], "Donald Duck")
        self.assertEqual(testdata["first_name"], "Donald")
        self.assertEqual(testdata["last_name"], "Duck")

    def test_login(self):
        self.do_login()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class OrbiOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.orbi.OrbiOAuth2"
    user_data_url = "https://login.orbi.kr/oauth/user/get"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
        }
    )
    user_data_body = json.dumps(
        {
            "username": "foobar",
            "first_name": "Foo",
            "last_name": "Bar",
            "name": "Foo Bar",
            "imin": "100000",
            "nick": "foobar",
            "photo": "http://s3.orbi.kr/data/member/wi/wizetdev_132894975780.jpeg",
            "sex": "M",
            "birth": "1973-08-03",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from .open_id_connect import OpenIdConnectAuth


class PixelPinOpenIDConnect(OpenIdConnectAuth):
    """PixelPin OpenID Connect authentication backend"""

    name = "pixelpin-openidconnect"
    ID_KEY = "sub"
    AUTHORIZATION_URL = "https://login.pixelpin.io/connect/authorize"
    ACCESS_TOKEN_URL = "https://login.pixelpin.io/connect/token"
    OIDC_ENDPOINT = "https://login.pixelpin.io"
    JWKS_URI = "https://login.pixelpin.io/.well-known/jwks"
    REQUIRES_EMAIL_VALIDATION = False

    def get_user_details(self, response):
        """Return user details from PixelPin account"""
        first_name = response.get("given_name")
        last_name = response.get("family_name")
        sub = response.get("sub")

        username = first_name + last_name + sub

        return {
            "username": username,
            "email": response.get("email"),
            "fullname": first_name + " " + last_name,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_json(
            "https://login.pixelpin.io/connect/userinfo",
            headers={"Authorization": f"Bearer {access_token}"},
        )



import requests
import responses

from ...utils import parse_qs
from .base import BaseBackendTest


class BaseLegacyTest(BaseBackendTest):
    form = ""
    response_body = ""

    def setUp(self):
        super().setUp()
        self.strategy.set_settings(
            {
                f"SOCIAL_AUTH_{self.name}_FORM_URL": self.strategy.build_absolute_uri(
                    f"/login/{self.backend.name}"
                )
            }
        )

    def extra_settings(self):
        return {f"SOCIAL_AUTH_{self.name}_FORM_URL": f"/login/{self.backend.name}"}

    def do_start(self):
        start_url = self.strategy.build_absolute_uri(self.backend.start().url)
        complete_url = self.complete_url
        assert complete_url, "Subclasses must set the complete_url attribute"

        responses.add(
            responses.GET,
            start_url,
            status=200,
            body=self.form.format(complete_url),
        )
        responses.add(
            responses.POST,
            complete_url,
            status=200,
            body=self.response_body,
            content_type="application/x-www-form-urlencoded",
        )
        response = requests.get(start_url, timeout=1)
        self.assertEqual(response.text, self.form.format(complete_url))
        response = requests.post(
            complete_url, data=parse_qs(self.response_body), timeout=1
        )
        self.strategy.set_request_data(parse_qs(response.text), self.backend)
        return self.backend.complete()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DisqusOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.disqus.DisqusOAuth2"
    user_data_url = "https://disqus.com/api/3.0/users/details.json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "code": 0,
            "response": {
                "username": "foobar",
                "numFollowers": 0,
                "isFollowing": False,
                "numFollowing": 0,
                "name": "Foo Bar",
                "numPosts": 0,
                "url": "",
                "isAnonymous": False,
                "rep": 1.231755,
                "about": "",
                "isFollowedBy": False,
                "connections": {},
                "emailHash": "5280f14cedf530b544aecc31fcfe0240",
                "reputation": 1.231755,
                "avatar": {
                    "small": {
                        "permalink": "https://disqus.com/api/users/avatars/foobar.jpg",
                        "cache": "https://securecdn.disqus.com/uploads/"
                        "users/453/4556/avatar32.jpg?1285535379",
                    },
                    "isCustom": False,
                    "permalink": "https://disqus.com/api/users/avatars/foobar.jpg",
                    "cache": "https://securecdn.disqus.com/uploads/users/453/"
                    "4556/avatar92.jpg?1285535379",
                    "large": {
                        "permalink": "https://disqus.com/api/users/avatars/foobar.jpg",
                        "cache": "https://securecdn.disqus.com/uploads/users/"
                        "453/4556/avatar92.jpg?1285535379",
                    },
                },
                "profileUrl": "http://disqus.com/foobar/",
                "numLikesReceived": 0,
                "isPrimary": True,
                "joinedAt": "2010-09-26T21:09:39",
                "id": "1010101",
                "location": "",
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""Tests for NGP VAN ActionID Backend"""

import datetime
from urllib.parse import urlencode

import pytest
import responses

from .open_id import OpenIdTest

JANRAIN_NONCE = datetime.datetime.now(datetime.timezone.utc).strftime(
    "%Y-%m-%dT%H:%M:%SZ"
)


class NGPVANActionIDOpenIDTest(OpenIdTest):
    """Test the NGP VAN ActionID OpenID 1.1 Backend"""

    backend_path = "social_core.backends.ngpvan.ActionIDOpenID"
    expected_username = "testuser@user.local"
    discovery_body = """<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns:openid="http://openid.net/xmlns/1.0" xmlns="xri://$xrd*($v*2.0)">
    <XRD>
        <Service priority="10">
            <Type>http://specs.openid.net/auth/2.0/signon</Type>
            <Type>http://openid.net/extensions/sreg/1.1</Type>
            <Type>http://axschema.org/contact/email</Type>
            <URI>https://accounts.ngpvan.com/OpenId/Provider</URI>
        </Service>
        <Service priority="20">
            <Type>http://openid.net/signon/1.0</Type>
            <Type>http://openid.net/extensions/sreg/1.1</Type>
            <Type>http://axschema.org/contact/email</Type>
            <URI>https://accounts.ngpvan.com/OpenId/Provider</URI>
        </Service>
    </XRD>
</xrds:XRDS>"""
    server_response = urlencode(
        {
            "openid.claimed_id": "https://accounts.ngpvan.com/user/abcd123",
            "openid.identity": "https://accounts.ngpvan.com/user/abcd123",
            "openid.sig": "Midw8F/rCDwW7vMz3y+vK6rjz6s=",
            "openid.signed": "claimed_id,identity,assoc_handle,op_endpoint,return_"
            "to,response_nonce,ns.alias3,alias3.mode,alias3.type."
            "alias1,alias3.value.alias1,alias3.type.alias2,alias3"
            ".value.alias2,alias3.type.alias3,alias3.value.alias3"
            ",alias3.type.alias4,alias3.value.alias4,alias3.type."
            "alias5,alias3.value.alias5,alias3.type.alias6,alias3"
            ".value.alias6,alias3.type.alias7,alias3.value.alias7"
            ",alias3.type.alias8,alias3.value.alias8,ns.sreg,sreg"
            ".fullname",
            "openid.assoc_handle": "{635790678917902781}{GdSyFA==}{20}",
            "openid.op_endpoint": "https://accounts.ngpvan.com/OpenId/Provider",
            "openid.return_to": "http://myapp.com/complete/actionid-openid/",
            "openid.response_nonce": JANRAIN_NONCE + "MMgBGEre",
            "openid.mode": "id_res",
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.ns.alias3": "http://openid.net/srv/ax/1.0",
            "openid.alias3.mode": "fetch_response",
            "openid.alias3.type.alias1": "http://openid.net/schema/contact/phone/b"
            "usiness",
            "openid.alias3.value.alias1": "+12015555555",
            "openid.alias3.type.alias2": "http://openid.net/schema/contact/interne"
            "t/email",
            "openid.alias3.value.alias2": "testuser@user.local",
            "openid.alias3.type.alias3": "http://openid.net/schema/namePerson/first",
            "openid.alias3.value.alias3": "John",
            "openid.alias3.type.alias4": "http://openid.net/schema/namePerson/last",
            "openid.alias3.value.alias4": "Smith",
            "openid.alias3.type.alias5": "http://axschema.org/namePerson/first",
            "openid.alias3.value.alias5": "John",
            "openid.alias3.type.alias6": "http://axschema.org/namePerson/last",
            "openid.alias3.value.alias6": "Smith",
            "openid.alias3.type.alias7": "http://axschema.org/namePerson",
            "openid.alias3.value.alias7": "John Smith",
            "openid.alias3.type.alias8": "http://openid.net/schema/namePerson",
            "openid.alias3.value.alias8": "John Smith",
            "openid.ns.sreg": "http://openid.net/extensions/sreg/1.1",
            "openid.sreg.fullname": "John Smith",
        }
    )

    def setUp(self):
        """Setup the test"""
        super().setUp()

        # Mock out the NGP VAN endpoints
        responses.add(
            responses.GET,
            "https://accounts.ngpvan.com/Home/Xrds",
            status=200,
            body=self.discovery_body,
        )
        responses.add(
            responses.GET,
            "https://accounts.ngpvan.com/user/abcd123",
            status=200,
            body=self.discovery_body,
        )
        responses.add(
            responses.GET,
            "https://accounts.ngpvan.com/OpenId/Provider",
            status=200,
            body=self.discovery_body,
        )

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_login(self):
        """Test the login flow using python-social-auth's built in test"""
        self.do_login()

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_partial_pipeline(self):
        """Test the partial flow using python-social-auth's built in test"""
        self.do_partial_pipeline()

    def test_get_ax_attributes(self):
        """Test that the AX attributes that NGP VAN responds with are present"""
        records = self.backend.get_ax_attributes()

        self.assertEqual(
            records,
            [
                ("http://openid.net/schema/contact/internet/email", "email"),
                ("http://openid.net/schema/contact/phone/business", "phone"),
                ("http://openid.net/schema/namePerson/first", "first_name"),
                ("http://openid.net/schema/namePerson/last", "last_name"),
                ("http://openid.net/schema/namePerson", "fullname"),
            ],
        )

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_setup_request(self):
        """Test the setup_request functionality in the NGP VAN backend"""
        # We can grab the requested attributes by grabbing the HTML of the
        # OpenID auth form and pulling out the hidden fields
        _, inputs = self.get_form_data(self.backend.auth_html())

        # Confirm that the only required attribute is email
        self.assertEqual(inputs["openid.ax.required"], "ngpvanemail")

        # Confirm that the 3 optional attributes are requested "if available"
        self.assertIn("ngpvanphone", inputs["openid.ax.if_available"])
        self.assertIn("ngpvanfirstname", inputs["openid.ax.if_available"])
        self.assertIn("ngpvanlastname", inputs["openid.ax.if_available"])

        # Verify the individual attribute properties
        self.assertEqual(
            inputs["openid.ax.type.ngpvanemail"],
            "http://openid.net/schema/contact/internet/email",
        )
        self.assertEqual(
            inputs["openid.ax.type.ngpvanfirstname"],
            "http://openid.net/schema/namePerson/first",
        )
        self.assertEqual(
            inputs["openid.ax.type.ngpvanlastname"],
            "http://openid.net/schema/namePerson/last",
        )
        self.assertEqual(
            inputs["openid.ax.type.ngpvanphone"],
            "http://openid.net/schema/contact/phone/business",
        )

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_user_data(self):
        """Ensure that the correct user data is being passed to create_user"""
        self.strategy.set_settings(
            {
                "USER_FIELDS": [
                    "email",
                    "first_name",
                    "last_name",
                    "username",
                    "phone",
                    "fullname",
                ]
            }
        )
        user = self.do_start()
        self.assertEqual(user.username, "testuser@user.local")
        self.assertEqual(user.email, "testuser@user.local")
        self.assertEqual(user.extra_user_fields["phone"], "+12015555555")
        self.assertEqual(user.extra_user_fields["first_name"], "John")
        self.assertEqual(user.extra_user_fields["last_name"], "Smith")
        self.assertEqual(user.extra_user_fields["fullname"], "John Smith")

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_extra_data_phone(self):
        """Confirm that you can get a phone number via the relevant setting"""
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_ACTIONID_OPENID_AX_EXTRA_DATA": [
                    ("http://openid.net/schema/contact/phone/business", "phone")
                ]
            }
        )
        user = self.do_start()
        self.assertEqual(user.social_user.extra_data["phone"], "+12015555555")

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_association_uid(self):
        """Test that the correct association uid is stored in the database"""
        user = self.do_start()
        self.assertEqual(
            user.social_user.uid, "https://accounts.ngpvan.com/user/abcd123"
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test
from .open_id_connect import OpenIdConnectTest


class TwitchOpenIdConnectTest(OpenIdConnectTest):
    backend_path = "social_core.backends.twitch.TwitchOpenIdConnect"
    user_data_url = "https://id.twitch.tv/oauth2/userinfo"
    issuer = "https://id.twitch.tv/oauth2"
    expected_username = "test_user1"
    openid_config_body = json.dumps(
        {
            "authorization_endpoint": "https://id.twitch.tv/oauth2/authorize",
            "claims_parameter_supported": True,
            "claims_supported": [
                "iss",
                "azp",
                "preferred_username",
                "updated_at",
                "aud",
                "exp",
                "iat",
                "picture",
                "sub",
                "email",
                "email_verified",
            ],
            "id_token_signing_alg_values_supported": [
                "RS256",
            ],
            "issuer": "https://id.twitch.tv/oauth2",
            "jwks_uri": "https://id.twitch.tv/oauth2/keys",
            "response_types_supported": [
                "id_token",
                "code",
                "token",
                "code id_token",
                "token id_token",
            ],
            "scopes_supported": [
                "openid",
            ],
            "subject_types_supported": [
                "public",
            ],
            "token_endpoint": "https://id.twitch.tv/oauth2/token",
            "token_endpoint_auth_methods_supported": [
                "client_secret_post",
            ],
            "userinfo_endpoint": "https://id.twitch.tv/oauth2/userinfo",
        }
    )


class TwitchOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.twitch.TwitchOAuth2"
    user_data_url = "https://api.twitch.tv/helix/users"
    expected_username = "test_user1"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
        }
    )
    user_data_body = json.dumps(
        {
            "data": [
                {
                    "id": "689563726",
                    "login": "test_user1",
                    "display_name": "test_user1",
                    "type": "",
                    "broadcaster_type": "",
                    "description": "",
                    "profile_image_url": "https://static-cdn.jtvnw.net/jtv_user_pictures/foo.png",
                    "offline_image_url": "",
                    "view_count": 0,
                    "email": "example@reply.com",
                    "created_at": "2021-05-21T18:59:25Z",
                    "access_token": "hmkgz15x7j54jm63rpwfwhcnue6t4fxwv",
                }
            ]
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class RedditOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.reddit.RedditOAuth2"
    user_data_url = "https://oauth.reddit.com/api/v1/me.json"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "name": "foobar",
            "created": 1203420772.0,
            "access_token": "foobar-token",
            "created_utc": 1203420772.0,
            "expires_in": 3600.0,
            "link_karma": 34,
            "token_type": "bearer",
            "comment_karma": 167,
            "over_18": True,
            "is_gold": False,
            "is_mod": True,
            "scope": "identity",
            "has_verified_email": False,
            "id": "33bma",
            "refresh_token": "foobar-refresh-token",
        }
    )
    user_data_body = json.dumps(
        {
            "name": "foobar",
            "created": 1203420772.0,
            "created_utc": 1203420772.0,
            "link_karma": 34,
            "comment_karma": 167,
            "over_18": True,
            "is_gold": False,
            "is_mod": True,
            "has_verified_email": False,
            "id": "33bma",
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 3600.0,
            "refresh_token": "foobar-new-refresh-token",
            "scope": "identity",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def refresh_token_arguments(self):
        uri = self.strategy.build_absolute_uri("/complete/reddit/")
        return {"redirect_uri": uri}

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")



"""
Facebook OAuth2, and Canvas Application  backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/facebook.html
"""

import base64
import hashlib
import hmac
import json
import time

from ..exceptions import (
    AuthCanceled,
    AuthException,
    AuthMissingParameter,
    AuthUnknownError,
)
from ..utils import constant_time_compare, handle_http_errors, parse_qs
from .oauth import BaseOAuth2

API_VERSION = 18.0


class FacebookOAuth2(BaseOAuth2):
    """Facebook OAuth2 authentication backend"""

    name = "facebook"
    REDIRECT_STATE = False
    RESPONSE_TYPE = None
    SCOPE_SEPARATOR = ","
    AUTHORIZATION_URL = "https://www.facebook.com/v{version}/dialog/oauth"
    ACCESS_TOKEN_URL = "https://graph.facebook.com/v{version}/oauth/access_token"
    ACCESS_TOKEN_METHOD = "GET"
    REVOKE_TOKEN_URL = "https://graph.facebook.com/v{version}/{uid}/permissions"
    REVOKE_TOKEN_METHOD = "DELETE"
    USER_DATA_URL = "https://graph.facebook.com/v{version}/me"
    EXTRA_DATA = [
        ("id", "id"),
        ("expires", "expires"),
        ("granted_scopes", "granted_scopes"),
        ("denied_scopes", "denied_scopes"),
    ]

    def auth_params(self, state=None):
        params = super().auth_params(state)
        params["return_scopes"] = "true"
        return params

    def authorization_url(self):
        version = self.setting("API_VERSION", API_VERSION)
        return self.AUTHORIZATION_URL.format(version=version)

    def access_token_url(self):
        version = self.setting("API_VERSION", API_VERSION)
        return self.ACCESS_TOKEN_URL.format(version=version)

    def get_user_details(self, response):
        """Return user details from Facebook account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("name", ""),
            response.get("first_name", ""),
            response.get("last_name", ""),
        )
        return {
            "username": response.get("username", response.get("name")),
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        params = self.setting("PROFILE_EXTRA_PARAMS", {}).copy()
        params["access_token"] = access_token

        if self.setting("APPSECRET_PROOF", True):
            _, secret = self.get_key_and_secret()
            params["appsecret_proof"] = hmac.new(
                secret.encode("utf8"),
                msg=access_token.encode("utf8"),
                digestmod=hashlib.sha256,
            ).hexdigest()

        version = self.setting("API_VERSION", API_VERSION)
        return self.get_json(self.USER_DATA_URL.format(version=version), params=params)

    def process_error(self, data):
        super().process_error(data)
        if data.get("error_code"):
            raise AuthCanceled(
                self, data.get("error_message") or data.get("error_code")
            )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        if not self.data.get("code"):
            raise AuthMissingParameter(self, "code")
        state = self.validate_state()
        key, secret = self.get_key_and_secret()
        response = self.request(
            self.access_token_url(),
            params={
                "client_id": key,
                "redirect_uri": self.get_redirect_uri(state),
                "client_secret": secret,
                "code": self.data["code"],
            },
            method=self.ACCESS_TOKEN_METHOD,
        )
        # API v2.3 returns a JSON, according to the documents linked at issue
        # #592, but it seems that this needs to be enabled(?), otherwise the
        # usual querystring type response is returned.
        try:
            response = response.json()
        except ValueError:
            response = parse_qs(response.text)
        access_token = response["access_token"]
        return self.do_auth(access_token, response, *args, **kwargs)

    def process_refresh_token_response(self, response, *args, **kwargs):
        try:
            return response.json()
        except ValueError:
            return parse_qs(response.content)

    def refresh_token_params(self, token, *args, **kwargs):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "fb_exchange_token": token,
            "grant_type": "fb_exchange_token",
            "client_id": client_id,
            "client_secret": client_secret,
        }

    @handle_http_errors
    def do_auth(self, access_token, response=None, *args, **kwargs):
        response = response or {}

        data = self.user_data(access_token)

        if not isinstance(data, dict):
            # From time to time Facebook responds back a JSON with just
            # False as value, the reason is still unknown, but since the
            # data is needed (it contains the user ID used to identify the
            # account on further logins), this app cannot allow it to
            # continue with the auth process.
            raise AuthUnknownError(
                self, "An error occurred while retrieving users Facebook data"
            )

        data["access_token"] = access_token
        if "expires_in" in response:
            data["expires"] = response["expires_in"]

        if self.data.get("granted_scopes"):
            data["granted_scopes"] = self.data["granted_scopes"].split(",")

        if self.data.get("denied_scopes"):
            data["denied_scopes"] = self.data["denied_scopes"].split(",")

        kwargs.update({"backend": self, "response": data})
        return self.strategy.authenticate(*args, **kwargs)

    def revoke_token_url(self, token, uid):
        version = self.setting("API_VERSION", API_VERSION)
        return self.REVOKE_TOKEN_URL.format(version=version, uid=uid)

    def revoke_token_params(self, token, uid):
        return {"access_token": token}

    def process_revoke_token_response(self, response):
        return (
            super().process_revoke_token_response(response)
            and response.content == "true"
        )


class FacebookAppOAuth2(FacebookOAuth2):
    """Facebook Application Authentication support"""

    name = "facebook-app"

    def uses_redirect(self):
        return False

    def auth_complete(self, *args, **kwargs):
        access_token = self.data.get("access_token")
        response = {}

        if "signed_request" in self.data:
            key, secret = self.get_key_and_secret()
            response = self.load_signed_request(self.data["signed_request"])
            assert response, "Missing signed_request response"
            if "user_id" not in response and "oauth_token" not in response:
                raise AuthException(self)

            if response is not None:
                access_token = (
                    response.get("access_token")
                    or response.get("oauth_token")
                    or self.data.get("access_token")
                )

        if access_token is None:
            if self.data.get("error") == "access_denied":
                raise AuthCanceled(self)
            raise AuthException(self)
        return self.do_auth(access_token, response, *args, **kwargs)

    def auth_html(self):
        key, secret = self.get_key_and_secret()
        namespace = self.setting("NAMESPACE", None)
        scope = self.setting("SCOPE", "")
        if scope:
            scope = self.SCOPE_SEPARATOR.join(scope)
        ctx = {
            "FACEBOOK_APP_NAMESPACE": namespace or key,
            "FACEBOOK_KEY": key,
            "FACEBOOK_EXTENDED_PERMISSIONS": scope,
            "FACEBOOK_COMPLETE_URI": self.redirect_uri,
        }
        tpl = self.setting("LOCAL_HTML", "facebook.html")
        return self.strategy.render_html(tpl=tpl, context=ctx)

    def load_signed_request(self, signed_request):
        def base64_url_decode(data):
            data = data.encode("ascii")
            data += b"=" * (4 - (len(data) % 4))
            return base64.urlsafe_b64decode(data)

        key, secret = self.get_key_and_secret()
        try:
            sig, payload = signed_request.split(".", 1)
        except ValueError:
            pass  # ignore if can't split on dot
        else:
            sig = base64_url_decode(sig)
            payload_json_bytes = base64_url_decode(payload)
            data = json.loads(payload_json_bytes.decode("utf-8", "replace"))
            expected_sig = hmac.new(
                secret.encode("ascii"),
                msg=payload.encode("ascii"),
                digestmod=hashlib.sha256,
            ).digest()
            # allow the signed_request to function for upto 1 day
            if constant_time_compare(sig, expected_sig) and data["issued_at"] > (
                time.time() - 86400
            ):
                return data



"""
Steam OpenId backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/steam.html
"""

from ..exceptions import AuthFailed
from .open_id import OpenIdAuth

USER_INFO = "http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?"


class SteamOpenId(OpenIdAuth):
    name = "steam"
    URL = "https://steamcommunity.com/openid"

    def get_user_id(self, details, response):
        """Return user unique id provided by service"""
        return self._user_id(response)

    def get_user_details(self, response):
        player = self.get_json(
            USER_INFO,
            params={
                "key": self.setting("API_KEY"),
                "steamids": self._user_id(response),
            },
        )
        if len(player["response"]["players"]) > 0:
            player = player["response"]["players"][0]
            details = {
                "username": player.get("personaname"),
                "email": "",
                "fullname": "",
                "first_name": "",
                "last_name": "",
                "player": player,
            }
        else:
            details = {}
        return details

    def consumer(self):
        # Steam seems to support stateless mode only, ignore store
        if not hasattr(self, "_consumer"):
            self._consumer = self.create_consumer()
        return self._consumer

    def _user_id(self, response):
        if not response.identity_url.startswith(self.URL):
            raise AuthFailed(self, "Openid identifier mismatch")
        user_id = response.identity_url.rsplit("/", 1)[-1]
        if not user_id.isdigit():
            raise AuthFailed(self, "Missing Steam Id")
        return user_id



"""
Backend for OpenID Connect EGI Check-in
https://www.egi.eu/service/check-in/
"""

from __future__ import annotations

from typing import Literal

from social_core.backends.open_id_connect import OpenIdConnectAuth

CHECKIN_ENV_ENDPOINTS = {
    "prod": "https://aai.egi.eu/auth/realms/egi",
    "demo": "https://aai-demo.egi.eu/auth/realms/egi",
    "dev": "https://aai-dev.egi.eu/auth/realms/egi",
}


class EGICheckinOpenIdConnect(OpenIdConnectAuth):
    name = "egi-checkin"
    # Check-in provides 3 environments: production, demo and development
    # Set the one to use as "prod", "demo" or "dev"
    CHECKIN_ENV: Literal["prod", "demo", "dev"] = "prod"
    # This is a opaque and unique id for every user that looks like an email
    # see https://docs.egi.eu/providers/check-in/sp/#1-community-user-identifier
    USERNAME_KEY = "voperson_id"
    EXTRA_DATA = [
        ("expires_in", "expires_in", True),
        ("refresh_token", "refresh_token", True),
        ("id_token", "id_token", True),
    ]
    # In order to get any scopes, you have to register your service with
    # Check-in, see documentation at https://docs.egi.eu/providers/check-in/sp/
    DEFAULT_SCOPE = [
        "openid",
        "profile",
        "email",
        "voperson_id",
        "eduperson_entitlement",
        "offline_access",
    ]
    # This is the list of entitlements that are allowed to login into the
    # service. A user with any of these will be allowed. If empty, all
    # users will be allowed
    ALLOWED_ENTITLEMENTS: list[str] = []

    def oidc_endpoint(self):
        endpoint = self.setting("OIDC_ENDPOINT", self.OIDC_ENDPOINT)
        if endpoint:
            return endpoint
        checkin_env = self.setting("CHECKIN_ENV", self.CHECKIN_ENV)
        return CHECKIN_ENV_ENDPOINTS.get(checkin_env, "")

    def get_user_details(self, response):
        username_key = self.setting("USERNAME_KEY", default=self.USERNAME_KEY)
        fullname, first_name, last_name = self.get_user_names(
            response.get("name") or "",
            response.get("given_name") or "",
            response.get("family_name") or "",
        )
        return {
            "username": response.get(username_key),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def entitlement_allowed(self, user_entitlements):
        allowed = True
        allowed_ent = self.setting("ALLOWED_ENTITLEMENTS", self.ALLOWED_ENTITLEMENTS)
        if allowed_ent:
            allowed = any(e in user_entitlements for e in allowed_ent)
        return allowed

    def auth_allowed(self, response, details):
        """Check-in promotes the use of eduperson_entitlements for AuthZ, if
        ALLOWED_ENTITLEMENTS is defined then use them to allow or not users"""
        allowed = super().auth_allowed(response, details)
        if allowed:
            user_entitlements = response.get("eduperson_entitlement") or []
            allowed = self.entitlement_allowed(user_entitlements)
        return allowed



"""
Tripit OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/tripit.html
"""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from defusedxml import minidom

from .oauth import BaseOAuth1

if TYPE_CHECKING:
    from xml.dom.minidom import Element


class TripItOAuth(BaseOAuth1):
    """TripIt OAuth authentication backend"""

    name = "tripit"
    AUTHORIZATION_URL = "https://www.tripit.com/oauth/authorize"
    REQUEST_TOKEN_URL = "https://api.tripit.com/oauth/request_token"
    ACCESS_TOKEN_URL = "https://api.tripit.com/oauth/access_token"
    EXTRA_DATA = [("screen_name", "screen_name")]

    def get_user_details(self, response):
        """Return user details from TripIt account"""
        fullname, first_name, last_name = self.get_user_names(response["name"])
        return {
            "username": response["screen_name"],
            "email": response["email"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        content: str = self.oauth_request(
            access_token, "https://api.tripit.com/v1/get/profile"
        ).text
        dom = minidom.parseString(content)
        profiles = dom.getElementsByTagName("Profile")
        public_display_names = dom.getElementsByTagName("public_display_name")
        screen_names = dom.getElementsByTagName("screen_name")
        primary_email = cast(
            "Element", dom.getElementsByTagName("is_primary")[0].parentNode
        ).getElementsByTagName("address")
        return {
            "id": profiles[0].getAttribute("ref"),
            "name": public_display_names[0].childNodes[0].nodeValue,
            "screen_name": screen_names[0].childNodes[0].nodeValue,
            "email": primary_email[0].childNodes[0].nodeValue,
        }



"""
Readability OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/readability.html
"""

from .oauth import BaseOAuth1

READABILITY_API = "https://www.readability.com/api/rest/v1"


class ReadabilityOAuth(BaseOAuth1):
    """Readability OAuth authentication backend"""

    name = "readability"
    ID_KEY = "username"
    AUTHORIZATION_URL = f"{READABILITY_API}/oauth/authorize/"
    REQUEST_TOKEN_URL = f"{READABILITY_API}/oauth/request_token/"
    ACCESS_TOKEN_URL = f"{READABILITY_API}/oauth/access_token/"
    EXTRA_DATA = [
        ("date_joined", "date_joined"),
        ("kindle_email_address", "kindle_email_address"),
        ("avatar_url", "avatar_url"),
        ("email_into_address", "email_into_address"),
    ]

    def get_user_details(self, response):
        fullname, first_name, last_name = self.get_user_names(
            first_name=response["first_name"], last_name=response["last_name"]
        )
        return {
            "username": response["username"],
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token):
        return self.get_json(
            READABILITY_API + "/users/_current", auth=self.oauth_auth(access_token)
        )



import json

from social_core.backends.orcid import ORCIDOAuth2

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ORCIDOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.orcid.ORCIDOAuth2"
    user_data_url = ORCIDOAuth2.USER_ID_URL
    expected_username = "0000-0002-2601-8132"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "orcid-identifier": {"path": "0000-0002-2601-8132"},
        }
    )
    user_data_body = json.dumps(
        {
            "orcid-identifier": {
                "uri": "http://orcid.org/0000-0002-2601-8132",
                "path": "0000-0002-2601-8132",
                "host": "orcid.org",
            },
            "person": {
                "last-modified-date": None,
                "name": {
                    "created-date": {"value": 1578249746904},
                    "last-modified-date": {"value": 1578249746904},
                    "given-names": {"value": "Janani Kantharooban"},
                    "family-name": {"value": "Umachanger"},
                    "credit-name": None,
                    "source": None,
                    "visibility": "PUBLIC",
                    "path": "0000-0002-2601-8132",
                },
                "other-names": {
                    "last-modified-date": None,
                    "other-name": [],
                    "path": "/0000-0002-2601-8132/other-names",
                },
                "biography": None,
                "researcher-urls": {
                    "last-modified-date": None,
                    "researcher-url": [],
                    "path": "/0000-0002-2601-8132/researcher-urls",
                },
                "emails": {
                    "last-modified-date": None,
                    "email": [],
                    "path": "/0000-0002-2601-8132/email",
                },
                "addresses": {
                    "last-modified-date": None,
                    "address": [],
                    "path": "/0000-0002-2601-8132/address",
                },
                "keywords": {
                    "last-modified-date": None,
                    "keyword": [],
                    "path": "/0000-0002-2601-8132/keywords",
                },
                "external-identifiers": {
                    "last-modified-date": None,
                    "external-identifier": [],
                    "path": "/0000-0002-2601-8132/external-identifiers",
                },
                "path": "/0000-0002-2601-8132/person",
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class EdmodoOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.edmodo.EdmodoOAuth2"
    user_data_url = "https://api.edmodo.com/users/me"
    expected_username = "foobar12345"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "username": "foobar12345",
            "coppa_verified": False,
            "first_name": "Foo",
            "last_name": "Bar",
            "premium": False,
            "verified_institution_member": False,
            "url": "https://api.edmodo.com/users/12345",
            "type": "teacher",
            "time_zone": None,
            "end_level": None,
            "start_level": None,
            "locale": "en",
            "subjects": None,
            "utc_offset": None,
            "email": "foo.bar@example.com",
            "gender": None,
            "about": None,
            "user_title": None,
            "id": 12345,
            "avatars": {
                "small": "https://api.edmodo.com/users/12345/avatar?type=small&u=5a15xug93m53mi4ey3ck4fvkq",
                "large": "https://api.edmodo.com/users/12345/avatar?type=large&u=5a15xug93m53mi4ey3ck4fvkq",
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from typing import Any
from urllib.parse import urlencode

from .oauth import BaseOAuth2


class UffdOAuth2(BaseOAuth2):
    """Uffd OAuth2 authentication backend

    You need to set the following config:
    SOCIAL_AUTH_UFFD_KEY - client id
    SOCIAL_AUTH_UFFD_SECRET - client secret
    SOCIAL_AUTH_UFFD_BASE_URL - base url to uffd installation
    """

    name = "uffd"
    REFRESH_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = " "
    STATE_PARAMETER = True
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("id", "id"),
    ]

    def get_user_details(self, response):
        """Return user details from a Uffd account"""
        fullname, first_name, last_name = self.get_user_names(
            fullname=response.get("name")
        )
        return {
            "username": response.get("nickname"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = self.userinfo_url() + "?" + urlencode({"access_token": access_token})
        try:
            user_data: dict[str, Any] = self.get_json(url)
        except ValueError:
            return None
        return user_data

    def authorization_url(self):
        return self.setting("BASE_URL") + "/oauth2/authorize"

    def access_token_url(self):
        return self.setting("BASE_URL") + "/oauth2/token"

    def userinfo_url(self):
        return self.setting("BASE_URL") + "/oauth2/userinfo"



"""
Vend  OAuth2 backend:
"""

from .oauth import BaseOAuth2


class VendOAuth2(BaseOAuth2):
    name = "vend"
    AUTHORIZATION_URL = "https://secure.vendhq.com/connect"
    ACCESS_TOKEN_URL = "https://{0}.vendhq.com/api/1.0/token"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token"),
        ("domain_prefix", "domain_prefix"),
    ]

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(self.data["domain_prefix"])

    def get_user_details(self, response):
        email = response["email"]
        username = response.get("username") or email.split("@", 1)[0]
        return {
            "username": username,
            "email": email,
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        prefix = kwargs["response"]["domain_prefix"]
        url = f"https://{prefix}.vendhq.com/api/users"
        data = self.get_json(url, headers={"Authorization": f"Bearer {access_token}"})
        return data["users"][0] if data.get("users") else {}



"""
Okta OAuth2 and OpenIdConnect:
    https://python-social-auth.readthedocs.io/en/latest/backends/okta.html
"""

from .okta import OktaOAuth2
from .open_id_connect import OpenIdConnectAuth


class OktaOpenIdConnect(OktaOAuth2, OpenIdConnectAuth):
    """Okta OpenID-Connect authentication backend"""

    name = "okta-openidconnect"
    REDIRECT_STATE = False
    RESPONSE_TYPE = "code"



DEFAULT_AUTH_PIPELINE = (
    # Get the information we can about the user and return it in a simple
    # format to create the user instance later. On some cases the details are
    # already part of the auth response from the provider, but sometimes this
    # could hit a provider API.
    "social_core.pipeline.social_auth.social_details",
    # Get the social uid from whichever service we're authing thru. The uid is
    # the unique identifier of the given user in the provider.
    "social_core.pipeline.social_auth.social_uid",
    # Verifies that the current auth process is valid within the current
    # project, this is where emails and domains whitelists are applied (if
    # defined).
    "social_core.pipeline.social_auth.auth_allowed",
    # Checks if the current social-account is already associated in the site.
    "social_core.pipeline.social_auth.social_user",
    # Make up a username for this person, appends a random string at the end if
    # there's any collision.
    "social_core.pipeline.user.get_username",
    # Send a validation email to the user to verify its email address.
    # 'social_core.pipeline.mail.mail_validation',
    # Associates the current social details with another user account with
    # a similar email address.
    # 'social_core.pipeline.social_auth.associate_by_email',
    # Create a user account if we haven't found one yet.
    "social_core.pipeline.user.create_user",
    # Create the record that associated the social account with this user.
    "social_core.pipeline.social_auth.associate_user",
    # Populate the extra_data field in the social record with the values
    # specified by settings (and the default ones like access_token, etc).
    "social_core.pipeline.social_auth.load_extra_data",
    # Update the user record with any changed info from the auth service.
    "social_core.pipeline.user.user_details",
)

DEFAULT_DISCONNECT_PIPELINE = (
    # Verifies that the social association can be disconnected from the current
    # user (ensure that the user login mechanism is not compromised by this
    # disconnection).
    "social_core.pipeline.disconnect.allowed_to_disconnect",
    # Collects the social associations to disconnect.
    "social_core.pipeline.disconnect.get_entries",
    # Revoke any access_token when possible.
    "social_core.pipeline.disconnect.revoke_tokens",
    # Removes the social associations.
    "social_core.pipeline.disconnect.disconnect",
)



# pyright: reportAttributeAccessIssue=false

import json

import responses

from .oauth import OAuth2PkcePlainTest, OAuth2PkceS256Test


class BitbucketDataCenterOAuth2Mixin:
    backend_path = "social_core.backends.bitbucket_datacenter.BitbucketDataCenterOAuth2"
    application_properties_url = (
        "https://bachmanity.atlassian.net/rest/api/latest/application-properties"
    )
    application_properties_headers = {"x-ausername": "erlich-bachman"}
    application_properties_body = json.dumps(
        {
            "version": "8.15.0",
            "buildNumber": "8015000",
            "buildDate": "1697764661289",
            "displayName": "Bitbucket",
        }
    )
    user_data_url = (
        "https://bachmanity.atlassian.net/rest/api/latest/users/erlich-bachman"
    )
    user_data_body = json.dumps(
        {
            "name": "erlich-bachman",
            "emailAddress": "erlich@bachmanity.com",
            "active": True,
            "displayName": "Erlich Bachman",
            "id": 1,
            "slug": "erlich-bachman",
            "type": "NORMAL",
            "links": {
                "self": [
                    {"href": "https://bachmanity.atlassian.net/users/erlich-bachman"}
                ]
            },
            "avatarUrl": "http://www.gravatar.com/avatar/af7d968fe79ea45271e3100391824b79.jpg?s=48&d=mm",
        }
    )
    access_token_body = json.dumps(
        {
            "scope": "PUBLIC_REPOS",
            "access_token": "dummy_access_token",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "dummy_refresh_token",
        }
    )
    refresh_token_body = json.dumps(
        {
            "scope": "PUBLIC_REPOS",
            "access_token": "dummy_access_token_refreshed",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "dummy_refresh_token_refreshed",
        }
    )
    expected_username = "erlich-bachman"

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {f"SOCIAL_AUTH_{self.name}_URL": "https://bachmanity.atlassian.net"}
        )
        return settings

    def auth_handlers(self, start_url):
        target_url = super().auth_handlers(start_url)
        responses.add(
            responses.GET,
            self.application_properties_url,
            body=self.application_properties_body,
            adding_headers=self.application_properties_headers,
            content_type="text/json",
        )
        return target_url

    def test_login(self):
        user = self.do_login()

        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "1")
        self.assertEqual(social.extra_data["first_name"], "Erlich")
        self.assertEqual(social.extra_data["last_name"], "Bachman")
        self.assertEqual(social.extra_data["email"], "erlich@bachmanity.com")
        self.assertEqual(social.extra_data["name"], "erlich-bachman")
        self.assertEqual(social.extra_data["username"], "erlich-bachman")
        self.assertEqual(social.extra_data["display_name"], "Erlich Bachman")
        self.assertEqual(social.extra_data["type"], "NORMAL")
        self.assertEqual(social.extra_data["active"], True)
        self.assertEqual(
            social.extra_data["url"],
            "https://bachmanity.atlassian.net/users/erlich-bachman",
        )
        self.assertEqual(
            social.extra_data["avatar_url"],
            "http://www.gravatar.com/avatar/af7d968fe79ea45271e3100391824b79.jpg?s=48&d=mm",
        )
        self.assertEqual(social.extra_data["scope"], "PUBLIC_REPOS")
        self.assertEqual(social.extra_data["access_token"], "dummy_access_token")
        self.assertEqual(social.extra_data["token_type"], "bearer")
        self.assertEqual(social.extra_data["expires"], 3600)
        self.assertEqual(social.extra_data["refresh_token"], "dummy_refresh_token")

    def test_refresh_token(self):
        _, social = self.do_refresh_token()

        self.assertEqual(social.uid, "1")
        self.assertEqual(social.extra_data["first_name"], "Erlich")
        self.assertEqual(social.extra_data["last_name"], "Bachman")
        self.assertEqual(social.extra_data["email"], "erlich@bachmanity.com")
        self.assertEqual(social.extra_data["name"], "erlich-bachman")
        self.assertEqual(social.extra_data["username"], "erlich-bachman")
        self.assertEqual(social.extra_data["display_name"], "Erlich Bachman")
        self.assertEqual(social.extra_data["type"], "NORMAL")
        self.assertEqual(social.extra_data["active"], True)
        self.assertEqual(
            social.extra_data["url"],
            "https://bachmanity.atlassian.net/users/erlich-bachman",
        )
        self.assertEqual(
            social.extra_data["avatar_url"],
            "http://www.gravatar.com/avatar/af7d968fe79ea45271e3100391824b79.jpg?s=48&d=mm",
        )
        self.assertEqual(social.extra_data["scope"], "PUBLIC_REPOS")
        self.assertEqual(
            social.extra_data["access_token"], "dummy_access_token_refreshed"
        )
        self.assertEqual(social.extra_data["token_type"], "bearer")
        self.assertEqual(social.extra_data["expires"], 3600)
        self.assertEqual(
            social.extra_data["refresh_token"], "dummy_refresh_token_refreshed"
        )


class BitbucketDataCenterOAuth2TestPkcePlain(
    BitbucketDataCenterOAuth2Mixin,
    OAuth2PkcePlainTest,
):
    pass


class BitbucketDataCenterOAuth2TestPkceS256(
    BitbucketDataCenterOAuth2Mixin,
    OAuth2PkceS256Test,
):
    pass



"""
Douban OAuth1 and OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/douban.html
"""

from .oauth import BaseOAuth1, BaseOAuth2


class DoubanOAuth(BaseOAuth1):
    """Douban OAuth authentication backend"""

    name = "douban"
    EXTRA_DATA = [("id", "id")]
    AUTHORIZATION_URL = "http://www.douban.com/service/auth/authorize"
    REQUEST_TOKEN_URL = "http://www.douban.com/service/auth/request_token"
    ACCESS_TOKEN_URL = "http://www.douban.com/service/auth/access_token"

    def get_user_id(self, details, response):
        return response["db:uid"]["$t"]

    def get_user_details(self, response):
        """Return user details from Douban"""
        return {"username": response["db:uid"]["$t"], "email": ""}

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "http://api.douban.com/people/%40me?&alt=json",
            auth=self.oauth_auth(access_token),
        )


class DoubanOAuth2(BaseOAuth2):
    """Douban OAuth authentication backend"""

    name = "douban-oauth2"
    AUTHORIZATION_URL = "https://www.douban.com/service/auth2/auth"
    ACCESS_TOKEN_URL = "https://www.douban.com/service/auth2/token"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("id", "id"),
        ("uid", "username"),
        ("refresh_token", "refresh_token"),
    ]

    def get_user_details(self, response):
        """Return user details from Douban"""
        fullname, first_name, last_name = self.get_user_names(response.get("name", ""))
        return {
            "username": response.get("uid", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data provided"""
        return self.get_json(
            "https://api.douban.com/v2/user/~me",
            headers={"Authorization": f"Bearer {access_token}"},
        )



from .oauth import BaseOAuth1


class WithingsOAuth(BaseOAuth1):
    name = "withings"
    AUTHORIZATION_URL = "https://developer.health.nokia.com/account/authorize"
    REQUEST_TOKEN_URL = "https://developer.health.nokia.com/account/request_token"
    ACCESS_TOKEN_URL = "https://developer.health.nokia.com/account/access_token"
    ID_KEY = "userid"

    def get_user_details(self, response):
        """Return user details from Withings account"""
        return {"userid": response["access_token"]["userid"], "email": ""}



"""
Shimmering Oauth
"""

from .oauth import BaseOAuth2


class ShimmeringOAuth2(BaseOAuth2):
    """Shimmering Verify OAuth2 authentication backend"""

    name = "shimmering"
    ID_KEY = "id"
    AUTHORIZATION_URL = "http://developers.shimmeringverify.com/o/authorize/"
    ACCESS_TOKEN_URL = "http://developers.shimmeringverify.com/o/token/"

    def get_user_details(self, response):
        """Return user details from Shimmering"""
        first_name = response.get("first_name")
        last_name = response.get("last_name")
        email = response.get("email")
        username = response.get("username")
        fullname = f"{first_name} {last_name}"
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {"Authorization": f"Bearer {access_token}"}
        return self.get_json(
            "http://developers.shimmeringverify.com/user_info/", headers=headers
        )



from __future__ import annotations

import unittest
from typing import Generic, TypeVar

import requests
import responses

from ...backends.base import BaseAuth
from ...backends.utils import load_backends, user_backends_data
from ...utils import PARTIAL_TOKEN_SESSION_NAME, module_member, parse_qs
from ..models import (
    TestAssociation,
    TestCode,
    TestNonce,
    TestStorage,
    TestUserSocialAuth,
    User,
)
from ..strategy import TestStrategy

BackendT = TypeVar("BackendT", bound=BaseAuth)


class BaseBackendTest(unittest.TestCase, Generic[BackendT]):
    backend: BackendT
    backend_path: str = ""
    name: str = ""
    complete_url = ""
    raw_complete_url = "/complete/{0}"
    expected_username: str = ""

    def setUp(self):
        responses.start()
        Backend = module_member(self.backend_path)
        self.strategy = TestStrategy(TestStorage)
        self.backend = Backend(self.strategy, redirect_uri=self.complete_url)
        self.name = self.backend.name.upper().replace("-", "_")
        self.complete_url = self.strategy.build_absolute_uri(
            self.raw_complete_url.format(self.backend.name)
        )
        backends = (
            self.backend_path,
            "social_core.tests.backends.test_broken.BrokenBackendAuth",
        )
        self.strategy.set_settings({"SOCIAL_AUTH_AUTHENTICATION_BACKENDS": backends})
        self.strategy.set_settings(self.extra_settings())
        # Force backends loading to trash PSA cache
        load_backends(backends, force_load=True)
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        TestCode.reset_cache()

    def tearDown(self):
        del self.backend
        self.strategy = None
        self.name = ""
        self.complete_url = None
        User.reset_cache()
        TestUserSocialAuth.reset_cache()
        TestNonce.reset_cache()
        TestAssociation.reset_cache()
        TestCode.reset_cache()
        responses.stop()
        responses.reset()

    def extra_settings(self) -> dict[str, str]:
        return {}

    def do_start(self):
        raise NotImplementedError("Implement in subclass")

    def do_login(self):
        user = self.do_start()
        username = self.expected_username
        self.assertEqual(user.username, username)
        self.assertEqual(self.strategy.session_get("username"), username)
        self.assertEqual(self.strategy.get_user(user.id), user)
        self.assertEqual(self.backend.get_user(user.id), user)
        user_backends = user_backends_data(
            user,
            self.strategy.get_setting("SOCIAL_AUTH_AUTHENTICATION_BACKENDS"),
            self.strategy.storage,
        )
        self.assertEqual(len(list(user_backends.keys())), 3)
        self.assertEqual("associated" in user_backends, True)
        self.assertEqual("not_associated" in user_backends, True)
        self.assertEqual("backends" in user_backends, True)
        self.assertEqual(len(user_backends["associated"]), 1)
        self.assertEqual(len(user_backends["not_associated"]), 1)
        self.assertEqual(len(user_backends["backends"]), 2)
        return user

    def pipeline_settings(self):
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PIPELINE": (
                    "social_core.pipeline.social_auth.social_details",
                    "social_core.pipeline.social_auth.social_uid",
                    "social_core.pipeline.social_auth.auth_allowed",
                    "social_core.tests.pipeline.ask_for_password",
                    "social_core.tests.pipeline.ask_for_slug",
                    "social_core.pipeline.social_auth.social_user",
                    "social_core.pipeline.user.get_username",
                    "social_core.pipeline.social_auth.associate_by_email",
                    "social_core.pipeline.user.create_user",
                    "social_core.pipeline.social_auth.associate_user",
                    "social_core.pipeline.social_auth.load_extra_data",
                    "social_core.tests.pipeline.set_password",
                    "social_core.tests.pipeline.set_slug",
                    "social_core.pipeline.user.user_details",
                )
            }
        )

    def pipeline_handlers(self, url):
        responses.add(responses.GET, url, status=200, body="foobar")
        responses.add(responses.POST, url, status=200)

    def pipeline_password_handling(self, url):
        password = "foobar"
        requests.get(url, timeout=1)
        requests.post(url, data={"password": password}, timeout=1)

        data = parse_qs(responses.calls[-1].request.body)
        self.assertEqual(data["password"], password)
        self.strategy.session_set("password", data["password"])
        return password

    def pipeline_slug_handling(self, url):
        slug = "foo-bar"
        requests.get(url, timeout=1)
        requests.post(url, data={"slug": slug}, timeout=1)

        data = parse_qs(responses.calls[-1].request.body)
        self.assertEqual(data["slug"], slug)
        self.strategy.session_set("slug", data["slug"])
        return slug

    def do_partial_pipeline(self):
        url = self.strategy.build_absolute_uri("/password")
        self.pipeline_settings()
        redirect = self.do_start()
        self.assertEqual(redirect.url, url)
        self.pipeline_handlers(url)

        password = self.pipeline_password_handling(url)
        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.assertEqual(partial.backend, self.backend.name)
        redirect = self.backend.continue_pipeline(partial)

        url = self.strategy.build_absolute_uri("/slug")
        self.assertEqual(redirect.url, url)
        self.pipeline_handlers(url)
        slug = self.pipeline_slug_handling(url)

        token = self.strategy.session_pop(PARTIAL_TOKEN_SESSION_NAME)
        partial = self.strategy.partial_load(token)
        self.assertEqual(partial.backend, self.backend.name)
        user = self.backend.continue_pipeline(partial)

        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(user.slug, slug)
        self.assertEqual(user.password, password)
        return user



"""
OAuth2 Backend to work with microsoft graph.
"""

import time

from ..exceptions import AuthMissingParameter
from .oauth import BaseOAuth2


class MicrosoftOAuth2(BaseOAuth2):
    name = "microsoft-graph"
    SCOPE_SEPARATOR = " "
    AUTHORIZATION_URL = "https://login.microsoftonline.com/common/oauth2/v2.0/authorize"
    ACCESS_TOKEN_URL = "https://login.microsoftonline.com/common/oauth2/v2.0/token"

    REDIRECT_STATE = False
    DEFAULT_SCOPE = ["User.Read"]

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        state = self.validate_state()

        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=self.auth_complete_credentials(),
            method=self.ACCESS_TOKEN_METHOD,
        )

        self.process_error(response)
        return self.do_auth(
            response["access_token"], response=response, *args, **kwargs
        )

    def get_user_id(self, details, response):
        """Use user account id as unique id"""
        return response.get("id")

    def get_user_details(self, response):
        """Return user details from Microsoft online account"""
        email = response.get("mail")
        username = response.get("userPrincipalName")

        if not username:
            raise AuthMissingParameter(self, "userPrincipalName")

        if "@" in username:
            if not email:
                email = username
            username = username.split("@", 1)[0]

        return {
            "username": username,
            "email": email,
            "fullname": response.get("displayName", ""),
            "first_name": response.get("givenName", ""),
            "last_name": response.get("surname", ""),
        }

    def user_data(self, access_token, *args, **kwargs):
        """Return user data by querying Microsoft service"""
        return self.get_json(
            "https://graph.microsoft.com/v1.0/me",
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
                "Authorization": "Bearer " + access_token,
            },
            method="GET",
        )

    def refresh_token_params(self, token, *args, **kwargs):
        return {
            "client_id": self.setting("KEY"),
            "client_secret": self.setting("SECRET"),
            "refresh_token": token,
            "grant_type": "refresh_token",
        }

    def get_auth_token(self, user_id):
        """Return the access token for the given user, after ensuring that it
        has not expired, or refreshing it if so."""
        user = self.get_user(user_id=user_id)
        access_token = user.social_user.access_token
        expires_on = user.social_user.extra_data["expires_on"]
        if expires_on <= int(time.time()):
            new_token_response = self.refresh_token(token=access_token)
            access_token = new_token_response["access_token"]
        return access_token



import json

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest


class GlobusOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.globus.GlobusOpenIdConnect"
    issuer = "https://auth.globus.org"
    openid_config_body = json.dumps(
        {
            "issuer": "https://auth.globus.org",
            "authorization_endpoint": "https://auth.globus.org/v2/oauth2/authorize",
            "userinfo_endpoint": "https://auth.globus.org/v2/oauth2/userinfo",
            "token_endpoint": "https://auth.globus.org/v2/oauth2/token",
            "revocation_endpoint": "https://auth.globus.org/v2/oauth2/token/revoke",
            "jwks_uri": "https://auth.globus.org/jwk.json",
            "response_types_supported": ["code", "token", "token id_token", "id_token"],
            "id_token_signing_alg_values_supported": ["RS512"],
            "scopes_supported": ["openid", "email", "profile"],
            "token_endpoint_auth_methods_supported": ["client_secret_basic"],
            "claims_supported": [
                "at_hash",
                "aud",
                "email",
                "exp",
                "name",
                "nonce",
                "preferred_username",
                "iat",
                "iss",
                "sub",
            ],
            "subject_types_supported": ["public"],
        }
    )



"""
Shopify OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/shopify.html
"""

import shopify

from ..exceptions import AuthCanceled, AuthFailed
from ..utils import handle_http_errors
from .oauth import BaseOAuth2


class ShopifyOAuth2(BaseOAuth2):
    """Shopify OAuth2 authentication backend"""

    name = "shopify"
    ID_KEY = "shop"
    EXTRA_DATA = [("shop", "shop"), ("website", "website"), ("expires", "expires")]
    REDIRECT_STATE = False

    @property
    def shopify_api_version(self):
        return self.setting("API_VERSION", "2020-10")

    def get_user_details(self, response):
        """Use the shopify store name as the username"""
        return {"username": str(response.get("shop", "")).replace(".myshopify.com", "")}

    def extra_data(self, user, uid, response, details=None, *args, **kwargs):
        """Return access_token and extra defined names to store in
        extra_data field"""
        data = super().extra_data(user, uid, response, details, *args, **kwargs)
        session = shopify.Session(
            self.data.get("shop").strip(), version=self.shopify_api_version
        )
        # Get, and store the permanent token
        token = session.request_token(data["access_token"])
        data["access_token"] = token
        return dict(data)

    def auth_url(self):
        key, secret = self.get_key_and_secret()
        shopify.Session.setup(api_key=key, secret=secret)
        scope = self.get_scope()
        state = self.state_token()
        self.strategy.session_set(self.name + "_state", state)
        redirect_uri = self.get_redirect_uri(state)
        session = shopify.Session(
            self.data.get("shop").strip(), version=self.shopify_api_version
        )
        return session.create_permission_url(scope=scope, redirect_uri=redirect_uri)

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        access_token = None
        key, secret = self.get_key_and_secret()
        try:
            shop_url = self.data.get("shop")
            shopify.Session.setup(api_key=key, secret=secret)
            shopify_session = shopify.Session(
                shop_url, version=self.shopify_api_version, token=self.data
            )
            access_token = shopify_session.token
        except shopify.ValidationException:
            raise AuthCanceled(self)
        else:
            if not access_token:
                raise AuthFailed(self, "Authentication Failed")
        return self.do_auth(
            access_token, shop_url, shopify_session.url, *args, **kwargs
        )

    def do_auth(self, access_token, shop_url, website, *args, **kwargs):
        kwargs.update(
            {
                "backend": self,
                "response": {
                    "shop": shop_url,
                    "website": f"http://{website}",
                    "access_token": access_token,
                },
            }
        )
        return self.strategy.authenticate(*args, **kwargs)



"""
Qiita OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/qiita.html
    http://qiita.com/api/v2/docs#get-apiv2oauthauthorize
    https://qiita.com/api/v2/docs#get-apiv2authenticated_user
"""

import json

from social_core.exceptions import AuthException

from .oauth import BaseOAuth2


class QiitaOAuth2(BaseOAuth2):
    """Qiita OAuth authentication backend"""

    name = "qiita"

    AUTHORIZATION_URL = "https://qiita.com/api/v2/oauth/authorize"
    ACCESS_TOKEN_URL = "https://qiita.com/api/v2/access_tokens"
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = True
    EXTRA_DATA = [
        ("description", "description"),
        ("facebook_id", "facebook_id"),
        ("followees_count", "followees_count"),
        ("followers_count", "followers_count"),
        ("github_login_name", "github_login_name"),
        ("id", "id"),
        ("items_count", "items_count"),
        ("linkedin_id", "linkedin_id"),
        ("location", "location"),
        ("name", "name"),
        ("organization", "organization"),
        ("permanent_id", "permanent_id"),
        ("profile_image_url", "profile_image_url"),
        ("team_only", "team_only"),
        ("twitter_screen_name", "twitter_screen_name"),
        ("website_url", "website_url"),
        ("image_monthly_upload_limit", "image_monthly_upload_limit"),
        ("image_monthly_upload_remaining", "image_monthly_upload_remaining"),
    ]

    # TODO: I am pretty sure this method returns the wrong type; it should
    # return a dict
    def auth_complete_params(self, state=None):  # type: ignore[reportIncompatibleMethodOverride]
        data = super().auth_complete_params(state)
        if "grant_type" in data:
            del data["grant_type"]
        if "redirect_uri" in data:
            del data["redirect_uri"]
        return json.dumps(data)

    def auth_headers(self):
        return {"Content-Type": "application/json"}

    def request_access_token(self, *args, **kwargs):
        data = super().request_access_token(*args, **kwargs)
        data.update({"access_token": data["token"]})
        return data

    def get_user_details(self, response):
        """Return user details from Qiita account"""
        return {
            "username": response["id"],
            "fullname": response["name"],
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://qiita.com/api/v2/authenticated_user",
            headers={"Authorization": f"Bearer {access_token}"},
        )

    def get_user_id(self, details, response):
        """Return user id"""
        user_id = None
        if self.setting("IDENTIFIED_BY_PERMANENT_ID"):
            user_id = response.get("permanent_id")
        else:
            user_id = response.get("id")

        if user_id is not None:
            return str(user_id)
        raise AuthException(self, "failed to get user id")



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class StocktwitsOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.stocktwits.StocktwitsOAuth2"
    user_data_url = "https://api.stocktwits.com/api/2/account/verify.json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "response": {"status": 200},
            "user": {
                "username": "foobar",
                "name": "Foo Bar",
                "classification": [],
                "avatar_url": "http://avatars.stocktwits.net/images/"
                "default_avatar_thumb.jpg",
                "avatar_url_ssl": "https://s3.amazonaws.com/st-avatars/images/"
                "default_avatar_thumb.jpg",
                "id": 101010,
                "identity": "User",
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()


class StocktwitsOAuth2UsernameAlternativeTest(StocktwitsOAuth2Test):
    user_data_body = json.dumps(
        {
            "response": {"status": 200},
            "user": {
                "username": "foobar",
                "name": "Foobar",
                "classification": [],
                "avatar_url": "http://avatars.stocktwits.net/images/"
                "default_avatar_thumb.jpg",
                "avatar_url_ssl": "https://s3.amazonaws.com/st-avatars/images/"
                "default_avatar_thumb.jpg",
                "id": 101010,
                "identity": "User",
            },
        }
    )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ChatworkOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.chatwork.ChatworkOAuth2"
    user_data_url = "https://api.chatwork.com/v2/me"
    expected_username = "hogehoge"
    access_token_body = json.dumps(
        {
            "access_token": "pyopyopyopyopyopyopyopyopyopyo",
            "token_type": "Bearer",
            "expires_in": "1501138041000",
            "refresh_token": "pyopyopyopyopyopyo",
            "scope": "rooms.all:read_write",
        }
    )

    user_data_body = json.dumps(
        {
            "account_id": 123,
            "room_id": 322,
            "name": "Foo Bar",
            "chatwork_id": "hogehoge",
            "organization_id": 101,
            "organization_name": "Foo foobar",
            "department": "Support",
            "title": "CMO",
            "url": "http://www.example.com",
            "introduction": "",
            "mail": "hogehoge@example.com",
            "tel_organization": "",
            "tel_extension": "",
            "tel_mobile": "",
            "skype": "",
            "facebook": "",
            "twitter": "",
            "avatar_image_url": "https://www.example.com/hogehoge.jpg",
            "login_mail": "hogehoge@example.com",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class OpenStreetMapOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.openstreetmap_oauth2.OpenStreetMapOAuth2"
    user_data_url = "https://api.openstreetmap.org/api/0.6/user/details.json"
    expected_username = "Steve"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "version": "0.6",
            "generator": "OpenStreetMap server",
            "copyright": "OpenStreetMap and contributors",
            "attribution": "http://www.openstreetmap.org/copyright",
            "license": "http://opendatacommons.org/licenses/odbl/1-0/",
            "user": {
                "id": 1,
                "display_name": "Steve",
                "account_created": "2005-09-13T15:32:57Z",
            },
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



from urllib.parse import quote

from .utils import (
    partial_pipeline_data,
    sanitize_redirect,
    setting_url,
    user_is_active,
    user_is_authenticated,
)


def do_auth(backend, redirect_name="next"):
    # Save any defined next value into session
    data = backend.strategy.request_data(merge=False)

    # Save extra data into session.
    for field_name in backend.setting("FIELDS_STORED_IN_SESSION", []):
        if field_name in data:
            backend.strategy.session_set(field_name, data[field_name])
        else:
            backend.strategy.session_set(field_name, None)

    if redirect_name in data:
        # Check and sanitize a user-defined GET/POST next field value
        redirect_uri = data[redirect_name]
        if backend.setting("SANITIZE_REDIRECTS", True):
            allowed_hosts = [
                *backend.setting("ALLOWED_REDIRECT_HOSTS", []),
                backend.strategy.request_host(),
            ]
            redirect_uri = sanitize_redirect(allowed_hosts, redirect_uri)
        backend.strategy.session_set(
            redirect_name, redirect_uri or backend.setting("LOGIN_REDIRECT_URL")
        )
    return backend.start()


def do_complete(backend, login, user=None, redirect_name="next", *args, **kwargs):
    data = backend.strategy.request_data()

    is_authenticated = user_is_authenticated(user)
    user = user if is_authenticated else None

    partial = partial_pipeline_data(backend, user, *args, **kwargs)
    if partial:
        user = backend.continue_pipeline(partial)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        user = backend.complete(user=user, redirect_name=redirect_name, *args, **kwargs)

    # pop redirect value before the session is trashed on login(), but after
    # the pipeline so that the pipeline can change the redirect if needed
    redirect_value = backend.strategy.session_get(redirect_name, "") or data.get(
        redirect_name, ""
    )

    # check if the output value is something else than a user and just
    # return it to the client
    user_model = backend.strategy.storage.user.user_model()
    if user and not isinstance(user, user_model):
        return user

    if is_authenticated:
        if not user:
            url = setting_url(backend, redirect_value, "LOGIN_REDIRECT_URL")
        else:
            url = setting_url(
                backend,
                redirect_value,
                "NEW_ASSOCIATION_REDIRECT_URL",
                "LOGIN_REDIRECT_URL",
            )
    elif user:
        if user_is_active(user):
            # catch is_new/social_user in case login() resets the instance
            is_new = getattr(user, "is_new", False)
            social_user = user.social_user
            login(backend, user, social_user)
            # store last login backend name in session
            backend.strategy.session_set(
                "social_auth_last_login_backend", social_user.provider
            )

            if is_new:
                url = setting_url(
                    backend,
                    "NEW_USER_REDIRECT_URL",
                    redirect_value,
                    "LOGIN_REDIRECT_URL",
                )
            else:
                url = setting_url(backend, redirect_value, "LOGIN_REDIRECT_URL")
        else:
            if backend.setting("INACTIVE_USER_LOGIN", False):
                social_user = user.social_user
                login(backend, user, social_user)
            url = setting_url(
                backend, "INACTIVE_USER_URL", "LOGIN_ERROR_URL", "LOGIN_URL"
            )
    else:
        url = setting_url(backend, "LOGIN_ERROR_URL", "LOGIN_URL")

    assert url, "By this point URL has to have been set"

    if redirect_value and redirect_value != url:
        redirect_value = quote(redirect_value)
        url += ("&" if "?" in url else "?") + f"{redirect_name}={redirect_value}"

    if backend.setting("SANITIZE_REDIRECTS", True):
        allowed_hosts = [
            *backend.setting("ALLOWED_REDIRECT_HOSTS", []),
            backend.strategy.request_host(),
        ]
        url = sanitize_redirect(allowed_hosts, url) or backend.setting(
            "LOGIN_REDIRECT_URL"
        )
    return backend.strategy.redirect(url)


def do_disconnect(
    backend, user, association_id=None, redirect_name="next", *args, **kwargs
):
    partial = partial_pipeline_data(backend, user, *args, **kwargs)
    if partial:
        if association_id and not partial.kwargs.get("association_id"):
            partial.extend_kwargs({"association_id": association_id})
        response = backend.disconnect(*partial.args, **partial.kwargs)
        # clean partial data after usage
        backend.strategy.clean_partial_pipeline(partial.token)
    else:
        response = backend.disconnect(
            user=user, association_id=association_id, *args, **kwargs
        )

    if isinstance(response, dict):
        url = backend.strategy.absolute_uri(
            backend.strategy.request_data().get(redirect_name, "")
            or backend.setting("DISCONNECT_REDIRECT_URL")
            or backend.setting("LOGIN_REDIRECT_URL")
        )
        if backend.setting("SANITIZE_REDIRECTS", True):
            allowed_hosts = [
                *backend.setting("ALLOWED_REDIRECT_HOSTS", []),
                backend.strategy.request_host(),
            ]
            url = (
                sanitize_redirect(allowed_hosts, url)
                or backend.setting("DISCONNECT_REDIRECT_URL")
                or backend.setting("LOGIN_REDIRECT_URL")
            )
        response = backend.strategy.redirect(url)
    return response



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MRGOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.mailru.MRGOAuth2"
    user_data_url = "https://oauth.mail.ru/userinfo"
    expected_username = "FooBar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "first_name": "Foo",
            "last_name": "Bar",
            "name": "Foo Bar",
            "locale": "ru_RU",
            "email": "foobar@example.com",
            "birthday": "11.07.1970",
            "gender": "m",
            "image": "http://cs7003.vk.me/v7003815/22a1/xgG9fb-IJ3Y.jpg",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
NationBuilder OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/nationbuilder.html
"""

from .oauth import BaseOAuth2


class NationBuilderOAuth2(BaseOAuth2):
    """NationBuilder OAuth2 authentication backend"""

    name = "nationbuilder"
    AUTHORIZATION_URL = "https://{slug}.nationbuilder.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://{slug}.nationbuilder.com/oauth/token"
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = [("id", "id"), ("expires", "expires")]

    def authorization_url(self):
        return self.AUTHORIZATION_URL.format(slug=self.slug)

    def access_token_url(self):
        return self.ACCESS_TOKEN_URL.format(slug=self.slug)

    @property
    def slug(self):
        return self.setting("SLUG")

    def get_user_details(self, response):
        """Return user details from Github account"""
        email = response.get("email") or ""
        username = email.split("@")[0] if email else ""
        return {
            "username": username,
            "email": email,
            "fullname": response.get("full_name") or "",
            "first_name": response.get("first_name") or "",
            "last_name": response.get("last_name") or "",
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        url = f"https://{self.slug}.nationbuilder.com/api/v1/people/me"
        return self.get_json(url, params={"access_token": access_token})["person"]



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class FiveHundredPxOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.five_hundred_px.FiveHundredPxOAuth"
    user_data_url = "https://api.500px.com/v1/users"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "user": {
                "id": 10101010,
                "username": "foobar",
                "firstname": "",
                "lastname": "",
                "birthday": None,
                "sex": 1,
                "city": "",
                "state": "",
                "country": "",
                "registration_date": "2011-11-11T00:00:00-00:00",
                "about": "",
                "usertype": 0,
                "domain": "UserName.500px.com",
                "fotomoto_on": False,
                "locale": "en",
                "show_nude": False,
                "allow_sale_requests": 1,
                "fullname": "UserName",
                "userpic_url": "https://graph.facebook.com/v2.7/"
                "1000000000/picture?height=100&width=100",
                "userpic_https_url": "https://graph.facebook.com/v2.7/"
                "1000000000/picture?"
                "height=100&width=100",
                "cover_url": None,
                "upgrade_status": 0,
                "store_on": False,
                "photos_count": 0,
                "galleries_count": 0,
                "affection": 51,
                "in_favorites_count": 0,
                "friends_count": 2,
                "followers_count": 3,
                "analytics_code": None,
                "invite_pending": False,
                "invite_accepted": False,
                "email": "user@user.com",
                "shadow_email": "user@user.com",
                "upload_limit": 20,
                "upload_limit_expiry": "2021-11-11T00:00:00-00:00",
                "upgrade_type": 0,
                "upgrade_status_expiry": "2011-11-21",
                "portfolio_enabled": False,
                "auth": {"facebook": 1, "twitter": 0, "google_oauth2": 1},
                "presubmit_for_licensing": None,
                "contacts": {"facebook": "1000000000"},
                "equipment": {},
                "avatars": {
                    "default": {
                        "http": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                        "https": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                    },
                    "large": {
                        "http": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                        "https": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                    },
                    "small": {
                        "http": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                        "https": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                    },
                    "tiny": {
                        "http": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                        "https": "https://graph.facebook.com/v2.7/"
                        "1000000000/picture?height=100&width=100",
                    },
                },
            }
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Patreon OAuth2 backend
https://www.patreon.com/platform/documentation/oauth
"""

from .oauth import BaseOAuth2


class PatreonOAuth2(BaseOAuth2):
    """Patreon OAuth2 authentication backend"""

    name = "patreon"
    AUTHORIZATION_URL = "https://www.patreon.com/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://www.patreon.com/api/oauth2/token"
    REVOKE_TOKEN_URL = "https://www.patreon.com/oauth2/revoke"
    REDIRECT_STATE = False
    ID_KEY = "id"
    EXTRA_DATA = [
        ("id", "id"),
    ]

    def get_user_details(self, response):
        details = response["attributes"]
        return {
            "username": details.get("full_name"),
            "email": details.get("email"),
            "fullname": details.get("full_name"),
            "first_name": details.get("first_name"),
            "last_name": details.get("last_name"),
        }

    def user_data(self, access_token, *args, **kwargs):
        return self.get_api(access_token, "identity")["data"]

    def get_api(self, access_token, suffix):
        return self.get_json(
            f"https://www.patreon.com/api/oauth2/v2/{suffix}",
            headers=self.get_auth_header(access_token),
        )

    def get_auth_header(self, access_token):
        return {"Authorization": f"Bearer {access_token}"}



import json

from social_core.exceptions import AuthException

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class QiitaOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.qiita.QiitaOAuth2"
    user_data_url = "https://qiita.com/api/v2/authenticated_user"
    expected_username = "foobar"

    access_token_body = json.dumps({"token": "foobar", "token_type": "bearer"})

    user_data_body = json.dumps(
        {
            "id": "foobar",
            "name": "Foo Bar",
            "description": "Hello, world.",
            "facebook_id": "qiita",
            "followees_count": 100,
            "followers_count": 200,
            "github_login_name": "qiitan",
            "items_count": 300,
            "linkedin_id": "qiita",
            "location": "Tokyo, Japan",
            "organization": "Qiita Inc.",
            "permanent_id": 12345,
            "profile_image_url": "https://s3-ap-northeast-1.amazonaws.com/qiita-image-store/0/88"
            "/ccf90b557a406157dbb9d2d7e543dae384dbb561/large.png?1575443439",
            "team_only": False,
            "twitter_screen_name": "qiita",
            "website_url": "https://qiita.com",
            "image_monthly_upload_limit": 1048576,
            "image_monthly_upload_remaining": 524288,
        }
    )

    def test_login(self):
        user = self.do_login()
        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "foobar")
        self.assertEqual(social.extra_data["permanent_id"], 12345)

    def test_partial_pipeline(self):
        user = self.do_partial_pipeline()
        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "foobar")
        self.assertEqual(social.extra_data["permanent_id"], 12345)


class QiitaOAuth2TestIdentifiedByPermanentId(QiitaOAuth2Test):
    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_QIITA_IDENTIFIED_BY_PERMANENT_ID": True}
        )

        user = self.do_login()
        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "12345")
        self.assertEqual(social.extra_data["permanent_id"], 12345)

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_QIITA_IDENTIFIED_BY_PERMANENT_ID": True}
        )

        user = self.do_partial_pipeline()
        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "12345")
        self.assertEqual(social.extra_data["permanent_id"], 12345)


class QiitaOAuth2TestIdentifiedByPermanentIdAuthException(QiitaOAuth2Test):
    user_data_body = json.dumps(
        {
            "id": "foobar",
            "name": "Foo Bar",
        }
    )

    def test_login(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_QIITA_IDENTIFIED_BY_PERMANENT_ID": True}
        )

        with self.assertRaises(AuthException):
            self.do_login()

    def test_partial_pipeline(self):
        self.strategy.set_settings(
            {"SOCIAL_AUTH_QIITA_IDENTIFIED_BY_PERMANENT_ID": True}
        )

        with self.assertRaises(AuthException):
            self.do_partial_pipeline()



from ..exceptions import AuthMissingParameter
from .base import BaseAuth


class LegacyAuth(BaseAuth):
    def get_user_id(self, details, response):
        return details.get(self.ID_KEY) or response.get(self.ID_KEY)

    def auth_url(self):
        return self.setting("FORM_URL")

    def auth_html(self):
        return self.strategy.render_html(tpl=self.setting("FORM_HTML"))

    def uses_redirect(self):
        return self.setting("FORM_URL") and not self.setting("FORM_HTML")

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        if self.ID_KEY not in self.data:
            raise AuthMissingParameter(self, self.ID_KEY)
        kwargs.update({"response": self.data, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)

    def get_user_details(self, response):
        """Return user details"""
        email = response.get("email", "")
        username = response.get("username", "")
        fullname, first_name, last_name = self.get_user_names(
            response.get("fullname", ""),
            response.get("first_name", ""),
            response.get("last_name", ""),
        )
        if email and not username:
            username = email.split("@", 1)[0]
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



# pyright: reportAttributeAccessIssue=false

import json

from social_core.exceptions import AuthException

from .oauth import (
    BaseAuthUrlTestMixin,
    OAuth2PkcePlainTest,
    OAuth2PkceS256Test,
    OAuth2Test,
)


class TwitterOAuth2Mixin:
    backend_path = "social_core.backends.twitter_oauth2.TwitterOAuth2"
    user_data_url = "https://api.twitter.com/2/users/me"
    access_token_body = json.dumps(
        {
            "token_type": "bearer",
            "expires_in": 7200,
            "access_token": "foobar",
            "scope": "users.read",
        }
    )
    user_data_body = json.dumps(
        {
            "data": {
                "id": "1234567890123456789",
                "username": "twitter_username",
                "name": "first last",
                "created_at": "2023-03-06T06:18:59.000Z",
                "public_metrics": {
                    "followers_count": 69,
                    "following_count": 129,
                    "tweet_count": 40,
                    "listed_count": 7,
                },
                "profile_image_url": "https://social-core-test-url.com/image.png",
                "verified_type": "none",
                "pinned_tweet_id": "9876543210987654321",
                "url": "https://social-core-test-url.com",
                "verified": False,
                "protected": True,
                "description": "description str",
                "entities": {
                    "url": {
                        "urls": [
                            {
                                "start": 0,
                                "end": 23,
                                "url": "entities-url-urls-url",
                                "expanded_url": "entities-url-urls-expanded_url",
                                "display_url": "entities-url-urls-display_url",
                            }
                        ]
                    },
                    "description": {
                        "urls": [
                            {
                                "start": 133,
                                "end": 156,
                                "url": "entities-description-urls-url",
                                "expanded_url": "entities-description-urls-expanded_url",
                                "display_url": "entities-description-urls-display_url",
                            }
                        ],
                        "hashtags": [
                            {
                                "start": 36,
                                "end": 44,
                                "tag": "entities-description-hashtags-tag",
                            }
                        ],
                    },
                },
            },
        },
    )

    expected_username = "twitter_username"

    def test_login(self):
        user = self.do_login()

        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "1234567890123456789")
        self.assertEqual(social.extra_data["first_name"], "first")
        self.assertEqual(social.extra_data["last_name"], "last")
        self.assertEqual(social.extra_data["fullname"], "first last")
        self.assertEqual(social.extra_data["created_at"], "2023-03-06T06:18:59.000Z")
        self.assertEqual(social.extra_data["verified"], False)
        self.assertEqual(social.extra_data["verified_type"], "none")
        self.assertEqual(social.extra_data["protected"], True)
        self.assertEqual(social.extra_data["description"], "description str")
        self.assertEqual(social.extra_data["url"], "https://social-core-test-url.com")
        self.assertEqual(social.extra_data["pinned_tweet_id"], "9876543210987654321")
        self.assertEqual(
            social.extra_data["profile_image_url"],
            "https://social-core-test-url.com/image.png",
        )
        self.assertEqual(social.extra_data["public_metrics"]["followers_count"], 69)
        self.assertEqual(social.extra_data["public_metrics"]["following_count"], 129)
        self.assertEqual(social.extra_data["public_metrics"]["tweet_count"], 40)
        self.assertEqual(social.extra_data["public_metrics"]["listed_count"], 7)

    def test_partial_pipeline(self):
        user = self.do_partial_pipeline()
        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "1234567890123456789")
        self.assertEqual(social.extra_data["first_name"], "first")
        self.assertEqual(social.extra_data["last_name"], "last")
        self.assertEqual(social.extra_data["fullname"], "first last")
        self.assertEqual(social.extra_data["created_at"], "2023-03-06T06:18:59.000Z")
        self.assertEqual(social.extra_data["verified"], False)
        self.assertEqual(social.extra_data["verified_type"], "none")
        self.assertEqual(social.extra_data["protected"], True)
        self.assertEqual(social.extra_data["description"], "description str")
        self.assertEqual(social.extra_data["url"], "https://social-core-test-url.com")
        self.assertEqual(social.extra_data["pinned_tweet_id"], "9876543210987654321")
        self.assertEqual(
            social.extra_data["profile_image_url"],
            "https://social-core-test-url.com/image.png",
        )
        self.assertEqual(social.extra_data["public_metrics"]["followers_count"], 69)
        self.assertEqual(social.extra_data["public_metrics"]["following_count"], 129)
        self.assertEqual(social.extra_data["public_metrics"]["tweet_count"], 40)
        self.assertEqual(social.extra_data["public_metrics"]["listed_count"], 7)


class TwitterOAuth2TestMissingOptionalValue(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.twitter_oauth2.TwitterOAuth2"
    user_data_url = "https://api.twitter.com/2/users/me"
    access_token_body = json.dumps(
        {
            "token_type": "bearer",
            "expires_in": 7200,
            "access_token": "foobar",
            "scope": "users.read",
        }
    )
    user_data_body = json.dumps(
        {
            "data": {
                "id": "1234567890123456789",
                "username": "twitter_username",
                "name": "first last",
            },
        },
    )

    expected_username = "twitter_username"

    def test_login(self):
        user = self.do_login()

        self.assertEqual(len(user.social), 1)

        social = user.social[0]
        self.assertEqual(social.uid, "1234567890123456789")
        self.assertEqual(social.extra_data["first_name"], "first")
        self.assertEqual(social.extra_data["last_name"], "last")
        self.assertEqual(social.extra_data["fullname"], "first last")

        self.assertIsNone(social.extra_data.get("created_at"))
        self.assertIsNone(social.extra_data.get("verified"))
        self.assertIsNone(social.extra_data.get("verified_type"))
        self.assertIsNone(social.extra_data.get("protected"))
        self.assertIsNone(social.extra_data.get("description"))
        self.assertIsNone(social.extra_data.get("url"))
        self.assertIsNone(social.extra_data.get("pinned_tweet_id"))
        self.assertIsNone(social.extra_data.get("profile_image_url"))
        self.assertIsNone(social.extra_data.get("public_metrics"))


class TwitterOAuth2TestPkcePlain(TwitterOAuth2Mixin, OAuth2PkcePlainTest):
    pass


class TwitterOAuth2TestPkceS256(TwitterOAuth2Mixin, OAuth2PkceS256Test):
    pass


class TwitterOAuth2TestInvalidCodeChallengeMethod(
    TwitterOAuth2Mixin, OAuth2PkcePlainTest
):
    def test_login__error(self):
        self.strategy.set_settings(
            {
                f"SOCIAL_AUTH_{self.name}_PKCE_CODE_CHALLENGE_METHOD": "invalidmethodname",
            }
        )

        with self.assertRaises(AuthException):
            self.do_login()



"""
Dribbble OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/dribbble.html
    http://developer.dribbble.com/v1/oauth/
"""

from .oauth import BaseOAuth2


class DribbbleOAuth2(BaseOAuth2):
    """Dribbble OAuth authentication backend"""

    name = "dribbble"
    AUTHORIZATION_URL = "https://dribbble.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://dribbble.com/oauth/token"
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = [
        ("id", "id"),
        ("name", "name"),
        ("html_url", "html_url"),
        ("avatar_url", "avatar_url"),
        ("bio", "bio"),
        ("location", "location"),
        ("links", "links"),
        ("buckets_count", "buckets_count"),
        ("comments_received_count", "comments_received_count"),
        ("followers_count", "followers_count"),
        ("followings_count", "followings_count"),
        ("likes_count", "likes_count"),
        ("likes_received_count", "likes_received_count"),
        ("projects_count", "projects_count"),
        ("rebounds_received_count", "rebounds_received_count"),
        ("shots_count", "shots_count"),
        ("teams_count", "teams_count"),
        ("pro", "pro"),
        ("buckets_url", "buckets_url"),
        ("followers_url", "followers_url"),
        ("following_url", "following_url"),
        ("likes_url", "shots_url"),
        ("teams_url", "teams_url"),
        ("created_at", "created_at"),
        ("updated_at", "updated_at"),
    ]

    def get_user_details(self, response):
        """Return user details from Dribbble account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        return {
            "username": response.get("username"),
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        return self.get_json(
            "https://api.dribbble.com/v1/user",
            headers={"Authorization": f"Bearer {access_token}"},
        )



"""
Chatwork OAuth2 backend
"""

import base64

from .oauth import BaseOAuth2


class ChatworkOAuth2(BaseOAuth2):
    """Chatwork OAuth authentication backend"""

    name = "chatwork"
    API_URL = "https://api.chatwork.com/v2"
    AUTHORIZATION_URL = "https://www.chatwork.com/packages/oauth2/login.php"
    ACCESS_TOKEN_URL = "https://oauth.chatwork.com/token"
    REDIRECT_STATE = True
    DEFAULT_SCOPE = ["users.profile.me:read"]
    ID_KEY = "account_id"
    EXTRA_DATA = [("expires_in", "expires"), ("refresh_token", "refresh_token")]

    def api_url(self, path):
        api_url = self.setting("API_URL") or self.API_URL
        return "{}{}".format(api_url.rstrip("/"), path)

    def auth_headers(self):
        return {
            "Authorization": b"Basic "
            + base64.b64encode("{}:{}".format(*self.get_key_and_secret()).encode())
        }

    def auth_complete_params(self, state=None):
        return {
            "grant_type": "authorization_code",
            "code": self.data.get("code", ""),
            "redirect_uri": self.get_redirect_uri(state),
        }

    def get_user_details(self, response):
        """Return user details from Chatwork account"""
        fullname, first_name, last_name = self.get_user_names(response.get("name"))
        username = (
            response.get("chatwork_id")
            or response.get("login_mail")
            or response.get("account_id")
        )
        email = response.get("mail") or response.get("login_mail") or ""
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data from service"""
        headers = {"Authorization": "Bearer " + access_token}
        return self.get_json(self.api_url("/me"), headers=headers)

    def refresh_token_params(self, token, *args, **kwargs):
        return {"refresh_token": token, "grant_type": "refresh_token"}



"""
Zotero OAuth1 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/zotero.html
"""

from .oauth import BaseOAuth1


class ZoteroOAuth(BaseOAuth1):
    """Zotero OAuth authorization mechanism"""

    name = "zotero"
    AUTHORIZATION_URL = "https://www.zotero.org/oauth/authorize"
    REQUEST_TOKEN_URL = "https://www.zotero.org/oauth/request"
    ACCESS_TOKEN_URL = "https://www.zotero.org/oauth/access"

    def get_user_id(self, details, response):
        """
        Return user unique id provided by service. For Ubuntu One
        the nickname should be original.
        """
        return details["userID"]

    def get_user_details(self, response):
        """Return user details from Zotero API account"""
        access_token = response.get("access_token", {})
        return {
            "username": access_token.get("username", ""),
            "userID": access_token.get("userID", ""),
        }



import hmac
import time
from base64 import urlsafe_b64decode, urlsafe_b64encode
from hashlib import sha256
from urllib.parse import urlencode

from ..exceptions import AuthException, AuthTokenError
from ..utils import parse_qs
from .base import BaseAuth


class DiscourseAuth(BaseAuth):
    name = "discourse"
    EXTRA_DATA = ["username", "name", "avatar_url"]

    def auth_url(self):
        """
        Get the URL to which we must redirect in order to authenticate the user
        """
        return_url = self.redirect_uri
        nonce = self.strategy.random_string(64)
        self.add_nonce(nonce)

        payload = urlencode({"nonce": nonce, "return_sso_url": return_url})
        base_64_payload = urlsafe_b64encode(payload.encode("utf8")).decode("ascii")

        payload_signature = hmac.new(
            self.setting("SECRET").encode("utf8"),
            base_64_payload.encode("utf8"),
            sha256,
        ).hexdigest()
        encoded_params = urlencode({"sso": base_64_payload, "sig": payload_signature})
        return f"{self.get_idp_url()}?{encoded_params}"

    def get_idp_url(self):
        return self.setting("SERVER_URL") + "/session/sso_provider"

    def get_user_id(self, details, response):
        return response["email"]

    def get_user_details(self, response):
        return {
            "username": response.get("username"),
            "email": response.get("email"),
            "name": response.get("name"),
            "groups": response.get("groups", "").split(","),
            "is_staff": response.get("admin") == "true"
            or response.get("moderator") == "true",
            "is_superuser": response.get("admin") == "true",
        }

    def add_nonce(self, nonce):
        self.strategy.storage.nonce.use(self.setting("SERVER_URL"), time.time(), nonce)

    def get_nonce(self, nonce):
        return self.strategy.storage.nonce.get(self.setting("SERVER_URL"), nonce)

    def delete_nonce(self, nonce):
        self.strategy.storage.nonce.delete(nonce)

    def auth_complete(self, *args, **kwargs):
        """
        The user has been redirected back from the IdP and we should
        now log them in, if everything checks out.
        """
        request_data = self.strategy.request_data()

        sso_params = request_data.get("sso")
        sso_signature = request_data.get("sig")

        param_signature = hmac.new(
            self.setting("SECRET").encode("utf8"), sso_params.encode("utf8"), sha256
        ).hexdigest()

        if not hmac.compare_digest(str(sso_signature), str(param_signature)):
            raise AuthException(self, "Could not verify discourse login")

        decoded_params = urlsafe_b64decode(sso_params.encode("utf8")).decode("ascii")

        # Validate the nonce to ensure the request was not modified
        response = parse_qs(decoded_params)
        nonce_obj = self.get_nonce(response.get("nonce"))
        if nonce_obj:
            self.delete_nonce(nonce_obj)
        else:
            raise AuthTokenError(self, "Incorrect id_token: nonce")

        kwargs.update({"sso": "", "sig": "", "backend": self, "response": response})
        return self.strategy.authenticate(*args, **kwargs)



import json

from social_core.backends.paypal import PayPalOAuth2

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class PayPalOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.paypal.PayPalOAuth2"
    user_data_url = (
        "https://api.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1"
    )
    expected_username = "mWq6_1sU85v5EG9yHdPxJRrhGHrnMJ-1PQKtX6pcsmA"
    access_token_body = json.dumps(
        {
            "token_type": "Bearer",
            "expires_in": 28800,
            "refresh_token": "foobar-refresh-token",
            "access_token": "foobar-token",
        }
    )
    user_data_body = json.dumps(
        {
            "user_id": "https://www.paypal.com/webapps/auth/identity/user/mWq6_1sU85v5EG9yHdPxJRrhGHrnMJ-1PQKtX6pcsmA",
            "name": "identity test",
            "given_name": "identity",
            "family_name": "test",
            "payer_id": "WDJJHEBZ4X2LY",
            "address": {
                "street_address": "1 Main St",
                "locality": "San Jose",
                "region": "CA",
                "postal_code": "95131",
                "country": "US",
            },
            "verified_account": True,
            "emails": [{"value": "user1@example.com", "primary": True}],
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "Bearer",
            "refresh_token": "foobar-new-refresh-token",
            "expires_in": 28800,
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()

    def test_refresh_token(self):
        user, social = self.do_refresh_token()
        self.assertEqual(user.username, self.expected_username)
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")

    def test_get_email_no_emails(self):
        emails = []
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, "")

    def test_get_email_multiple_emails(self):
        expected_email = "mail2@example.com"
        emails = [
            {"value": "mail1@example.com", "primary": False},
            {"value": expected_email, "primary": True},
        ]
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, expected_email)

    def test_get_email_multiple_emails_no_primary(self):
        expected_email = "mail1@example.com"
        emails = [
            {"value": expected_email, "primary": False},
            {"value": "mail2@example.com", "primary": False},
        ]
        email = PayPalOAuth2.get_email(emails)
        self.assertEqual(email, expected_email)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SketchfabOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.sketchfab.SketchfabOAuth2"
    user_data_url = "https://sketchfab.com/v2/users/me"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "uid": "42",
            "email": "foo@bar.com",
            "displayName": "foo bar",
            "username": "foobar",
            "apiToken": "XXX",
        }
    )

    def test_login(self):
        self.do_login()

    def test_partial_pipeline(self):
        self.do_partial_pipeline()



"""
Created on May 13, 2014

@author: Yong Zhang (zyfyfe@gmail.com)
"""

import json

from ..utils import parse_qs
from .oauth import BaseOAuth2


class QQOAuth2(BaseOAuth2):
    name = "qq"
    ID_KEY = "openid"
    AUTHORIZE_URL = "https://graph.qq.com/oauth2.0/authorize"
    ACCESS_TOKEN_URL = "https://graph.qq.com/oauth2.0/token"
    AUTHORIZATION_URL = "https://graph.qq.com/oauth2.0/authorize"
    OPENID_URL = "https://graph.qq.com/oauth2.0/me"
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("nickname", "username"),
        ("figureurl_qq_1", "profile_image_url"),
        ("gender", "gender"),
    ]

    def get_user_details(self, response):
        """
        Return user detail from QQ account sometimes nickname will duplicate
        with another qq account, to avoid this issue it's possible to use
        openid as username.
        """
        if self.setting("USE_OPENID_AS_USERNAME", False):
            username = response.get("openid", "")
        else:
            username = response.get("nickname", "")

        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("nickname", "")
        )

        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def get_openid(self, access_token):
        response = self.request(self.OPENID_URL, params={"access_token": access_token})
        content = response.content.decode()
        data = json.loads(content[10:-3])
        return data["openid"]

    def user_data(self, access_token, *args, **kwargs):
        openid = self.get_openid(access_token)
        response = self.get_json(
            "https://graph.qq.com/user/get_user_info",
            params={
                "access_token": access_token,
                "oauth_consumer_key": self.setting("KEY"),
                "openid": openid,
            },
        )
        response["openid"] = openid
        return response

    def request_access_token(self, url, data, *args, **kwargs):
        response = self.request(url, *args, **kwargs)
        return parse_qs(response.content)



"""
HubSpot OAuth2 backend, docs at:
    https://developers.hubspot.com/docs/methods/oauth2/oauth2-overview
"""

from .oauth import BaseOAuth2


class HubSpotOAuth2(BaseOAuth2):
    """HubSpot OAuth2 authentication backend"""

    name = "hubspot"
    AUTHORIZATION_URL = "https://app.hubspot.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.hubapi.com/oauth/v1/token"
    USER_DATA_URL = "https://api.hubapi.com/oauth/v1/access-tokens/"
    DEFAULT_SCOPE = ["oauth"]
    EXTRA_DATA = [
        ("hub_domain", "hub_domain"),
        ("hub_id", "hub_id"),
        ("app_id", "app_id"),
        ("user_id", "user_id"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details"""
        response["email"] = response["user"]
        return response

    def user_data(self, access_token, *args, **kwargs):
        """Loads user data information from service"""
        return self.get_json(
            self.USER_DATA_URL + access_token,
            headers={"Authorization": "Bearer " + access_token},
        )
